from typing import List, Dict

import casadi as ca
import numpy as np
from itertools import accumulate

from coker.backends.backend import get_backend_by_name


from coker.dynamics import (
    VariationalProblem,
    split_at_non_differentiable_points,
    ControlVariable,
    ParameterVariable,
    ConstantControlVariable,
    SpikeVariable,
    BoundedVariable,
    PiecewiseConstantVariable,
    InterpolatingPoly,
    InterpolatingPolyCollection,
    VariationalSolution,
)


def noop(*args):
    return None


def create_variational_solver(problem: VariationalProblem):

    # At each point t_i in segment j
    # ([D_{ij}(t_i)] * 1_{n})  X_j - (dt) f(t_i, X_{ij}, Z_{ij}, U(t_i), p) = 0
    #                                     g(t_i, X_{ij}, Z_{ij}, U(t_i), p) = 0
    # ([D_{ij}(t_i)] * 1_{n})  Q_j - (dt) h(t_i, X_{ij}, Z_{ij}, U(t_i), p) = 0

    # If U(t) is discontinuous, it is evaluated as if from below at t_i = 1
    # and from above at t_i = -1

    # At the end of each segment, we also have the constraint
    # [X, Z, Q]_{j - 1}(1) = [X, Z, Q]_{j}(-1),

    # with the [X]_{-1} = [X0(0, U(0), p)]
    #                 0 = g(0, X0(0,u,p), Z, U(0))
    #            Q_{-1} = 0

    # cost = cost(y(t, x, z, q, u, p), u, p)
    # calling cost symbolically, we should know the set of $t_i$ that are
    # required

    casadi = get_backend_by_name("casadi")

    x_dim, z_dim, q_dim = problem.system.get_state_dimensions()
    x_size = x_dim.flat()
    z_size = z_dim.flat() if z_dim else 0
    q_size = q_dim.flat() if q_dim else 0

    tolerance = problem.transcription_options.absolute_tolerance

    # initial intervals
    intervals = split_at_non_differentiable_points(
        problem.control if problem.control else [],
        problem.t_final,
        problem.transcription_options,
    )

    colocation_points = [problem.transcription_options.minimum_degree] * len(
        intervals
    )
    poly_collection = SymbolicPolyCollection(
        name="x",
        dimension=x_size + z_size + q_size,
        intervals=intervals,
        degrees=colocation_points,
    )
    time_points, _, _ = zip(*list(poly_collection.knot_points()))

    proj_x = ca.hcat([ca.MX.eye(x_size), ca.MX.zeros(q_size, z_size)])
    proj_z = ca.hcat(
        [
            ca.MX.zeros(z_size, x_size),
            ca.MX.eye(z_size),
            ca.MX.zeros(z_size, q_size),
        ]
    )
    proj_q = ca.hcat(
        [
            ca.MX.zeros(q_size, x_size),
            ca.MX.zeros(q_size, z_size),
            ca.MX.eye(q_size),
        ]
    )

    # Set up functions
    equalities = []

    dynamics = lambda *args: casadi.evaluate(problem.system.dxdt, args)
    algebraic = lambda *args: (
        casadi.evaluate(problem.system.g, args) if problem.system.g else noop
    )
    quadrature = lambda *args: (
        casadi.evaluate(problem.system.dqdt, args)
        if problem.system.dqdt
        else noop
    )

    # parameters & control - need to know how big they are
    # to construct our decision variables
    control_variables, parameters = problem.control, problem.parameters

    p, p_symbols, p0_guess, (p_lower, p_guess, p_upper), p_output_map = (
        construct_parameters(parameters)
    )
    if problem.system_parameter_map is not None:
        proj_p = ca.DM(problem.system_parameter_map)
    else:
        proj_p = ca.DM.eye(p.shape[0])

    control_factory = (
        ControlFactory(control_variables, problem.t_final)
        if control_variables
        else noop
    )

    if problem.control:
        u_symbols = control_factory.symbols()
        u_lower, u_upper = (
            control_factory.lower_bounds,
            control_factory.upper_bounds,
        )
        u_guess = control_factory.guess
    else:
        u_symbols = ca.MX.zeros(0)
        u_lower, u_upper = [], []
        u_guess = noop

    (t0, x0_symbol), *x_start = list(poly_collection.interval_starts())
    x_end = list(poly_collection.interval_ends())[:-1]

    x0_guess, z0_guess = casadi.evaluate(
        problem.system.x0, [0, u_guess, proj_p @ p0_guess]
    )
    x0_val, z0_val = casadi.evaluate(
        problem.system.x0, [0, control_factory, proj_p @ p]
    )

    equalities += [
        proj_x @ x0_symbol - x0_val,
    ]
    if z_size > 0:
        equalities.append(proj_z @ z0_val)

    equalities += [
        xs_i - xe_i for ((_, xs_i), (_, xe_i)) in zip(x_start, x_end)
    ]
    for poly in poly_collection.polys:
        interval_dynamics = []
        interval_quadratures = []
        for t, v, dv in poly.knot_points():
            x = proj_x @ v
            z = proj_z @ v
            if z_size > 0:
                z += z0_val

            dx = proj_x @ dv

            (dynamics_ij,) = dynamics(t, x, z, control_factory, proj_p @ p)
            interval_dynamics.append(dynamics_ij)
            equalities.append(dx - dynamics_ij)

            if q_size > 0:
                dq = proj_q @ dv
                (quadrature_ij,) = quadrature(
                    t, x, z, control_factory, proj_p @ p
                )
                equalities.append(dq - quadrature_ij)
                interval_quadratures.append(quadrature_ij)

            if z_size > 0:
                (alg,) = algebraic(t, x, z, control_factory, proj_p @ p)
                equalities.append(alg)

            # xend = x_start + int_tstart^t_end f(x)dt
            # ->  0 = xend - xstart - [dx_0 | dx_1 | dx_2 | ...] @ w

        _, vstart = poly.start_point()
        _, vend = poly.end_point()

        dX = ca.hcat(interval_dynamics)
        assert poly.weights[0, poly.degree] == 0
        w = ca.DM(poly.weights[0, :-1])

        equalities.append(proj_x @ vend - proj_x @ vstart - (dX @ w))
        if q_size > 0:
            dQ = ca.hcat(interval_quadratures)
            equalities.append(proj_q @ vend - proj_q @ vstart - dQ @ w)

    path_symbols = poly_collection.symbols()

    decision_variables = ca.vertcat(path_symbols, u_symbols, p_symbols)

    lower_bound = ca.vertcat(
        -ca.DM.ones(poly_collection.size()) * ca.inf, *u_lower, p_lower
    )

    upper_bound = ca.vertcat(
        ca.DM.ones(poly_collection.size()) * ca.inf, *u_upper, p_upper
    )

    # cost
    #
    def solution_proxy(*args):
        if problem.control:
            tau, u_val, p_val = args
        else:
            tau, p_val = args
            u_val = ca.MX.zeros(u_symbols.shape)
        inner = poly_collection(tau)
        x_tau = proj_x @ inner
        z_tau = proj_z @ inner
        q_tau = proj_q @ inner
        (y_val,) = casadi.evaluate(
            problem.system.y, [tau, x_tau, z_tau, u_val, proj_p @ p_val, q_tau]
        )
        return y_val

    if problem.control:
        (cost,) = casadi.evaluate(
            problem.loss, [solution_proxy, control_factory, p]
        )
    else:
        (cost,) = casadi.evaluate(problem.loss, [solution_proxy, p])

    g = ca.vertcat(*[e for e in equalities if e is not None])
    ubg = tolerance * ca.DM.ones(g.shape)
    lbg = -tolerance * ca.DM.ones(g.shape)

    t_end, v_end = poly_collection.polys[-1].end_point()
    x_end = proj_x @ v_end
    z_end = proj_z @ v_end
    q_end = proj_q @ v_end
    u_end = control_factory(t_end, x_end, z_end)
    end_args = (t_end, x_end, z_end, u_end, p, q_end)

    for constraint in problem.terminal_constraints:
        (g_inner,) = casadi.evaluate(constraint.value, end_args)
        g_lower = casadi.to_backend_array(constraint.lower)
        g_upper = casadi.to_backend_array(constraint.upper)
        assert g_lower.shape == g_inner.shape == g_inner.shape
        g = ca.vertcat(g, g_inner)
        lbg = ca.vertcat(lbg, g_lower)
        ubg = ca.vertcat(ubg, g_upper)

    solver_options = problem.transcription_options.optimiser_options

    if not problem.transcription_options.verbose:
        solver_options.update(
            {
                "ipopt.print_level": 0,
                "print_time": False,
                "ipopt.sb": "yes",
            }
        )
    u0_guess = ca.DM.zeros(u_symbols.shape)

    state_guess = ca.vertcat(
        x0_guess,
        z0_guess if z0_guess is not None else ca.DM.zeros(z_size),
        ca.DM.zeros(q_size),
    )

    n_reps = int(path_symbols.shape[0] / state_guess.shape[0])

    decision_variables_0 = ca.vertcat(
        ca.repmat(state_guess, n_reps), u0_guess, p_guess
    )

    if problem.transcription_options.initialise_near_guess:

        init_spec = {
            "f": cost,
            "x": decision_variables,
            "g": ca.vertcat(g, p_symbols, u_symbols),
        }
        init_solver = ca.nlpsol(
            "initialiser", "ipopt", init_spec, solver_options
        )

        init_soln = init_solver(
            x0=decision_variables_0,
            lbx=lower_bound,
            ubx=upper_bound,
            lbg=ca.vertcat(lbg, p_guess, ca.DM.zeros(u_symbols.shape)),
            ubg=ca.vertcat(ubg, p_guess, ca.DM.zeros(u_symbols.shape)),
        )
        decision_variables_0 = init_soln["x"]
        initial_cost = float(init_soln["f"])
        assert (
            initial_cost <= ca.inf
        ), f"Cost at guess {initial_cost} is not finite"

    nlp_spec = {"f": cost, "x": decision_variables, "g": g}
    nlp_solver = ca.nlpsol("solver", "ipopt", nlp_spec, solver_options)
    soln = nlp_solver(
        x0=decision_variables_0,
        lbx=lower_bound,
        ubx=upper_bound,
        lbg=lbg,
        ubg=ubg,
    )

    min_loss = float(soln["f"])
    min_args = soln["x"]

    f_out = ca.Function(
        "Output",
        [decision_variables],
        [path_symbols, u_symbols, p],
        {},
    )
    x_out, u_out, p_out = f_out(min_args)

    path = poly_collection.to_fixed(np.array(x_out))
    projectors = tuple(
        (
            np.array(proj.to_DM()).reshape(proj.shape)
            if proj.shape != (0, 1)
            else None
        )
        for proj in (proj_x, proj_z, proj_q)
    )
    p_out = np.array(p_out)
    parameter_out = p_output_map(p_out)

    control_out = (
        control_factory.to_output_array(u_out) if problem.control else None
    )

    solution = VariationalSolution(
        cost=min_loss,
        projectors=projectors,
        parameter_solutions=parameter_out,
        parameters=p_out.flatten(),
        path=path,
        control_solutions=control_out,
        output=problem.system.y,
    )

    return solution


class SymbolicPoly(InterpolatingPoly):
    def __init__(self, name, dimension, interval, degree):
        size = (degree + 1) * dimension
        values = ca.MX.sym(name, size)
        super().__init__(dimension, interval, degree, values)

    def symbols(self):
        return self.values

    def __call__(self, t):
        s = self._interval_to_s(t)
        try:
            i = next(i for i, s_i in enumerate(self.s) if abs(s_i - s) < 1e-9)
            return self.values[i * self.dimension : (i + 1) * self.dimension]
        except (StopIteration, TypeError):
            pass
        n = len(self.s)
        s_vector = ca.vertcat(*[s**i for i in range(n)])

        projection = s_vector.T @ ca.DM(self.bases)

        # Casadi's reshape follows fortran convention (unlike numpy),
        # so we need to transpose the operation
        value = ca.reshape(self.values, (self.dimension, -1)) @ projection.T

        return ca.reshape(value, (self.dimension, 1))


class SymbolicPolyCollection(InterpolatingPolyCollection):
    def symbols(self):
        return ca.vertcat(*[p.values for p in self.polys])

    def __init__(self, name, dimension, intervals, degrees):
        assert len(intervals) == len(degrees)
        polys = [
            SymbolicPoly(f"{name}_{i}", dimension, interval, degree)
            for i, (interval, degree) in enumerate(zip(intervals, degrees))
        ]
        super().__init__(polys)

    def to_fixed(self, array):
        size = sum(p.size() for p in self.polys)
        np_array = np.array(array)
        assert np_array.shape == (size, 1)
        np_array.reshape((size,))

        slices = []
        offset = 0
        for p in self.polys:
            slices.append(slice(offset, offset + p.size()))
            offset += p.size()

        polys = [
            InterpolatingPoly(p.dimension, p.interval, p.degree, np_array[slc])
            for (p, slc) in zip(self.polys, slices)
        ]

        return InterpolatingPolyCollection(polys)

    def __call__(self, t):
        if isinstance(t, (ca.SX, ca.MX)):
            result = 0
            for i, (start, end) in enumerate(self.intervals):
                poly_eval = self.polys[i](t)
                factor_1 = ca.if_else(t > start, poly_eval, 0)
                factor_2 = ca.if_else(t < end, factor_1, 0)
                result += factor_2
            return result
        return super().__call__(t)


class ParameterOutputMap:
    def __init__(self, indices: Dict[str, int]):
        self.indices = indices

    def __call__(self, value: ca.DM):
        return {name: float(value[i, 0]) for name, i in self.indices.items()}


def construct_parameters(parameters: List[ParameterVariable]):

    params = []
    upper_bounds = []
    guess = []
    lower_bounds = []
    symbols = {}
    p0 = []
    output_map = {}
    for i, p in enumerate(parameters):
        if isinstance(p, BoundedVariable):
            try:
                symbol = symbols[p.name]
                params.append(symbol)
                index = output_map[p.name]
                p0.append(p0[index])
                continue
            except KeyError:
                pass
            symbol = ca.MX.sym(f"{p.name}")
            output_map[p.name] = len(params)
            params.append(symbol)
            symbols[p.name] = symbol
            upper_bounds.append(
                p.upper_bound if p.upper_bound is not None else ca.inf
            )
            guess.append(p.guess)
            p0.append(p.guess)
            lower_bounds.append(
                p.lower_bound if p.lower_bound is not None else -ca.inf
            )
        elif isinstance(p, (float, int)):
            params.append(ca.MX(p))
            p0.append(p)
        else:
            raise ValueError(f"Parameter {p} is not a valid parameter")

    params = ca.vertcat(*params)

    symbols = ca.vertcat(*symbols.values())
    return (
        params,  # actual parameter vector
        symbols,  # symbols
        ca.DM(p0),  # actual parameter vector guess
        (
            ca.DM(lower_bounds),
            ca.DM(guess),
            ca.DM(upper_bounds),
        ),  # symbols bounds and guess
        ParameterOutputMap(output_map),  # map to symbols
    )


class ControlPath:
    def __init__(self, vector: ca.MX, rate: float):
        self.vector = vector
        self.rate = rate

    def __call__(self, t) -> ca.MX:
        index = int(t * self.rate)
        return self.vector[index]


class ControlFactory:
    def __init__(self, variables: List[ControlVariable], t_final: float):
        self.t_final = t_final
        self.variables = variables
        self._symbols = [
            ca.MX.sym(v.name, v.degrees_of_freedom(0, t_final))
            for v in variables
        ]
        self.values = []
        self.upper_bounds = [
            ca.DM.ones(v.degrees_of_freedom(0, t_final))
            * (v.upper_bound if v.upper_bound != np.inf else ca.inf)
            for v in variables
        ]
        self.lower_bounds = [
            ca.DM.ones(v.degrees_of_freedom(0, t_final))
            * (v.lower_bound if v.lower_bound != -np.inf else -ca.inf)
            for v in variables
        ]
        self.sizes = [v.degrees_of_freedom(0, t_final) for v in variables]
        self.offsets = accumulate(self.sizes)

    def guess(self, _):
        return ca.DM.zeros(len(self.variables))

    def symbols(self) -> ca.MX:
        return ca.vertcat(*self._symbols)

    def __call__(self, t):
        assert (
            0 <= t <= self.t_final
        ), f"Control variable is not defined at t = {t}"
        out = []
        for s, var in zip(self._symbols, self.variables):
            if isinstance(var, ConstantControlVariable):
                out.append(s)
            elif isinstance(var, SpikeVariable):
                out.append(s if abs(t - var.time) < 1e-9 else 0)
            elif isinstance(var, PiecewiseConstantVariable):
                index = int(t * var.sample_rate)
                out.append(s[index])
            else:
                raise ValueError(
                    f"Control variable {var} is not a valid control variable"
                )
        return ca.vertcat(*out)

    def to_output_array(self, solution: ca.DM):
        return [
            v.to_solution(solution[offset : offset + size])
            for v, offset, size in zip(
                self.variables, self.offsets, self.sizes
            )
        ]



import sympy as sp
import numpy as np
from coker import Function
from coker.backends.backend import Backend, ArrayLike
from coker.algebra.ops import OP, ConcatenateOP, ReshapeOP, NormOP
from coker.algebra.dimensions import Dimension
from coker.backends.numpy.core import reshape

MatrixType = (sp.Matrix, sp.ImmutableMatrix)


def sympy_mul(x, y):

    if isinstance(x, MatrixType) and isinstance(y, MatrixType):
        return x.multiply_elementwise(y)
    result = x * y
    return result


def to_matrix(x):
    if isinstance(x, sp.ImmutableMatrix):
        return x
    if isinstance(x, sp.Matrix):
        return sp.ImmutableMatrix(x)
    if isinstance(x, sp.Array):
        return sp.ImmutableMatrix(x.tomatrix())
    if isinstance(x, list):
        array = sp.Array(x)
        if len(array.shape) == 1:
            array.reshape(*array.shape, 1)
            return to_matrix(array)
        if len(array.shape) == 2:
            return to_matrix(array)
        raise ValueError("Cannot convert tensor to matrix")

    raise NotImplementedError(f"Unknown type: {x}")


def sympy_div(x, y):
    if isinstance(x, MatrixType) and isinstance(y, MatrixType):
        y_inv = sp.zeros(*y.shape)
        for i in range(y.shape[0]):
            for j in range(y.shape[1]):
                y_inv[i, j] = 1 / y[i, j]
        return sp.ImmutableMatrix(x.multiply_elementwise(y_inv))
    return x / y


def sympy_matmul(x, y):
    if isinstance(x, sp.Matrix) and isinstance(y, sp.Matrix):
        return to_matrix(to_matrix(x) * to_matrix(y))

    idx = len(x.shape)
    assert x.shape[idx - 1] == y.shape[0]
    result = sp.tensorcontraction(sp.tensorproduct(x, y), (idx - 1, idx))
    assert result.shape == (*x.shape[0:-1], *y.shape[1:])

    if len(result.shape) == 2:
        return to_matrix(result)
    return result


def sympy_dot(x, y):
    if isinstance(x, MatrixType) and isinstance(y, MatrixType):
        return to_matrix(x).dot(to_matrix(y))
    try:
        return sp.tensorcontraction(sp.tensorproduct(x, y), (0, 1))
    except:
        pass
    return x.T @ y


impls = {
    OP.ADD: lambda x, y: x + y,
    OP.SUB: lambda x, y: x - y,
    OP.MUL: sympy_mul,
    OP.DIV: sympy_div,
    OP.MATMUL: sympy_matmul,
    OP.SIN: sp.sin,
    OP.COS: sp.cos,
    OP.TAN: sp.tan,
    OP.EXP: sp.exp,
    OP.PWR: lambda x, y: x**y,
    OP.INT_PWR: lambda x, y: x**y,
    OP.ARCCOS: sp.acos,
    OP.ARCSIN: sp.asin,
    OP.DOT: sympy_dot,
    OP.CROSS: lambda x, y: x.cross(y),
    OP.TRANSPOSE: sp.transpose,
    OP.NEG: lambda x: -x,
    OP.SQRT: sp.sqrt,
    OP.ABS: lambda x: sp.Abs(x),
    OP.EQUAL: lambda x, y: x == y,
    OP.CASE: lambda cond, t, f: t if cond else f,
    OP.ARCTAN2: sp.atan2,
    OP.EVALUATE: lambda op, *args: op(*args),
    OP.LOG: sp.log,
}


def sympy_concat(*arrays, axis: int = 0):

    if axis == 0:
        return sp.Matrix.vstack(*arrays)
    if axis == 1:
        return sp.Matrix.hstack(*arrays)
    raise NotImplementedError


parameterised_impls = {
    ConcatenateOP: lambda op, *x: sympy_concat(*x, axis=op.axis),
    ReshapeOP: lambda op, x: reshape(x, dim=Dimension(op.newshape)),
    #    NormOP: lambda op, x:
}


class SympyBackend(Backend):

    def to_numpy_array(self, array):

        if isinstance(array, np.ndarray) and array.dtype in {
            float,
            np.float64,
        }:
            return array
        if isinstance(array, (int, float, complex)):
            return array

        try:
            if not array.is_constant():
                return array
        except AttributeError:
            pass

        try:
            value = sp.nsimplify(array, tolerance=1e-10)
            out = np.array(value, dtype=float)
            return out
        except (AttributeError, TypeError):
            pass

        raise ValueError(f"Cannot convert {array} to a numpy array")

    def to_backend_array(self, array):
        import scipy.sparse

        if scipy.sparse.issparse(array):
            array = array.toarray()
        if isinstance(array, np.ndarray):
            if len(array.shape) == 1:
                return to_matrix(
                    sp.Array(array.tolist(), shape=(array.shape[0], 1))
                )
            elif len(array.shape) == 2:
                return to_matrix(sp.Array(array.tolist(), shape=array.shape))
            return sp.Array(array.tolist(), shape=array.shape)
        if isinstance(array, list):
            try:
                return to_matrix(array)
            except ValueError:
                return sp.Array(array)
        if isinstance(array, np.float64):
            return sp.Float(float(array))
        if isinstance(
            array, (sp.ImmutableDenseNDimArray, sp.MutableDenseNDimArray)
        ):
            if len(array.shape) == 1:
                array = array.reshape(*array.shape, 1)
            if len(array.shape) == 2:
                return to_matrix(array)

        return array

    def native_types(self):
        return [sp.Array, sp.Symbol, sp.Float, sp.Integer, sp.Rational]

    def reshape(self, array, shape):
        result = reshape(array, shape)
        try:
            if len(result.shape) == 2:
                return to_matrix(result)
        except (AttributeError, ValueError):
            pass
        return self.to_backend_array(result)

    def call(self, op, *args):
        if op in impls:
            result = impls[op](*args)
            return result

        if isinstance(op, tuple(parameterised_impls.keys())):
            kls = op.__class__

            result = parameterised_impls[kls](op, *args)
            return result

        raise NotImplementedError(f"{op} is not implemented")

    def evaluate(self, function: Function, inputs: ArrayLike):

        results = super().evaluate(function, inputs)

        def eval(x):
            if x is None:
                return None
            if isinstance(x, np.ndarray):
                return x
            try:
                if x.is_constant():
                    return self.to_numpy_array(x)
            except (AttributeError, ValueError):
                pass

            try:
                return sp.nsimplify(x, tolerance=1e-10)
            except (TypeError, ValueError):
                pass
            if isinstance(x, sp.ImmutableDenseNDimArray):
                return x.applyfunc(eval)
            return x

        output = []
        for result, shape in zip(results, function.output_shape()):
            if shape is None:
                output.append(result)
                continue
            result = eval(result)

            if (
                result is not None
                and not shape.is_scalar()
                and result.shape != shape.shape
                and isinstance(result, MatrixType)
            ):
                result = sp.Array(result)
                result = result.reshape(*shape)
                output.append(result)

            else:
                output.append(result)

        return output

    def lower_to_symbolic(self, function: Function):
        """Return (args, output) as sympy symbolic expressions.

        This is a sympy-specific utility for inspecting or printing functions
        symbolically; it is separate from lower() which returns a callable.
        """
        from coker.backends.evaluator import evaluate_inner

        tape = function.tape
        args = []
        workspace = {}
        for idx, name in zip(tape.input_indicies, tape.input_names):
            if idx < 0:
                args.append(None)
                continue
            dim = tape.dim[idx]
            if dim.is_scalar():
                sym = sp.Symbol(name)
            else:
                shape = dim.shape
                if len(shape) == 1:
                    sym = sp.Array(
                        [sp.Symbol(f"{name}_{i}") for i in range(shape[0])]
                    )
                else:
                    sym = sp.Array(
                        [
                            [
                                sp.Symbol(f"{name}_{i}_{j}")
                                for j in range(shape[1])
                            ]
                            for i in range(shape[0])
                        ]
                    )
            args.append(sym)
            workspace[idx] = sym
        outputs = evaluate_inner(tape, args, function.output, self, workspace)
        if function.is_single:
            return args, outputs[0]
        return args, outputs

    def build_optimisation_problem(*args):
        raise NotImplementedError("not supported on sympy backend")

    def evaluate_integrals(*args):

        raise NotImplementedError("not supported on sympy backend")



import dataclasses
from functools import reduce
from operator import mul
from typing import List, Optional, Tuple, Union


@dataclasses.dataclass
class VectorSpace:
    name: str
    dimension: Union[int, Tuple[int, ...]]

    @property
    def size(self) -> int:
        if isinstance(self.dimension, int):
            return self.dimension
        return reduce(mul, self.dimension)


@dataclasses.dataclass
class Scalar:
    name: str

    @property
    def size(self) -> int:
        return 1


class Dimension:
    def __init__(self, tuple_or_none):

        if isinstance(tuple_or_none, int):
            tuple_or_none = (tuple_or_none,)

        assert tuple_or_none is None or isinstance(tuple_or_none, tuple)
        self.dim = tuple_or_none

    def flat(self):
        d = 1
        if self.dim is not None:
            for d_i in self.dim:
                d *= d_i
        return d

    def to_space(self, name) -> Union[Scalar, VectorSpace]:
        if self.is_scalar():
            return Scalar(name)
        return VectorSpace(name, self.dim)

    def index_iterator(self, row_major=False):

        if not self.dim:
            return (0,)

        count = 1
        for d in self.dim:
            count *= d

        # Compute per-dimension strides for mixed-radix decomposition.
        # column-major (row_major=False): first index varies fastest → strides [1, d0, d0*d1, ...]
        # row-major (row_major=True):    last  index varies fastest → strides [..., d1, 1]
        strides = []
        stride = 1
        if row_major:
            for d in reversed(self.dim):
                strides.insert(0, stride)
                stride *= d
        else:
            for d in self.dim:
                strides.append(stride)
                stride *= d

        for i in range(count):
            yield tuple((i // s) % d for s, d in zip(strides, self.dim))

    def __eq__(self, other):
        return self.dim == other.dim

    def is_scalar(self):
        return self.dim is None

    def is_vector(self):
        return self.dim is not None and len(self.dim) == 1

    def is_covector(self):
        return not self.is_scalar() and len(self.dim) == 2 and self.dim[0] == 1

    def is_matrix(self):
        return not self.is_scalar() and len(self.dim) == 2 and self.dim[0] > 1

    def is_multilinear_map(self):
        return isinstance(self.dim, tuple) and len(self.dim) > 2

    def __iter__(self):
        return iter(self.dim)

    def __repr__(self):
        if self.dim is None:
            return "R"

        return repr(self.dim)

    @property
    def shape(self):
        if self.dim is None:
            return (1,)
        return self.dim


@dataclasses.dataclass
class FunctionSpace:
    """Represents a function space, including its domain and codomain.

    Attributes:
        name (str): The name of the function space.
        arguments (List[Scalar | VectorSpace]): A list specifying the input arguments of
            the function, where each argument can be either a scalar or a vector space.
        output (List[Scalar | VectorSpace]): A list specifying the output of the function,
            where each element can be either a scalar or a vector space.
        signature (Optional[Tuple[int]]): Optional list of integers denoting the degree
            of differentiability for each argument. If not provided, the default is
            smooth (infinitely differentiable) functions.

    """

    name: str
    arguments: List[Scalar | VectorSpace]
    output: List[Scalar | VectorSpace]
    signature: Optional[Tuple[int]] = None
    """Optional list of integers, specifying the degree of differentiability for each argument. Defaults to infinite (i.e. smooth function)."""

    def input_dimensions(self):
        return [
            (
                Dimension(None)
                if isinstance(arg, Scalar)
                else Dimension(arg.dimension)
            )
            for arg in self.arguments
        ]

    def output_dimensions(self):
        if self.output is None:
            return (None,)
        return [
            (
                Dimension(None)
                if isinstance(out, Scalar)
                else Dimension(out.dimension)
            )
            for out in self.output
        ]

    def is_scalar(self):
        return (
            len(self.output_dimensions()) == 1
            and self.output_dimensions()[0].is_scalar()
        )


@dataclasses.dataclass
class Element:
    parent: VectorSpace



from typing import Type, Tuple, List

import numpy as np
import jax.numpy as jnp

from coker.algebra import Dimension, OP
from coker.algebra.kernel import Tracer
from coker.algebra.ops import ConcatenateOP, ReshapeOP, NormOP

from coker.backends.backend import Backend, ArrayLike


def to_array(value, shape):

    if isinstance(value, np.ndarray) and value.shape == shape:
        return jnp.array(value)

    raise plementedError


scalar_types = (
    jnp.float32,
    jnp.float64,
    np.float64,
    np.float32,
    np.int32,
    np.int64,
    jnp.int32,
    jnp.int64,
    float,
    complex,
    int,
    bool,
    jnp.bool_,
)


def div(num, den):
    if (num == 0).all() and (den == 0):
        return num
    else:
        return jnp.divide(num, den)


impls = {
    OP.ADD: jnp.add,
    OP.SUB: jnp.subtract,
    OP.MUL: jnp.multiply,
    OP.DIV: div,
    OP.MATMUL: jnp.matmul,
    OP.SIN: jnp.sin,
    OP.COS: jnp.cos,
    OP.TAN: jnp.tan,
    OP.EXP: jnp.exp,
    OP.PWR: jnp.power,
    OP.INT_PWR: jnp.power,
    OP.ARCCOS: jnp.arccos,
    OP.ARCSIN: jnp.arcsin,
    OP.DOT: jnp.dot,
    OP.CROSS: jnp.cross,
    OP.TRANSPOSE: jnp.transpose,
    OP.NEG: jnp.negative,
    OP.SQRT: jnp.sqrt,
    OP.ABS: jnp.abs,
    OP.ARCTAN2: jnp.arctan2,
    OP.LESS_EQUAL: jnp.less_equal,
    OP.LESS_THAN: jnp.less,
    OP.EQUAL: jnp.equal,
    OP.CASE: lambda c, t, f: t if c else f,
    OP.EVALUATE: lambda op, *args: op(*args),
    OP.LOG: jnp.log,
}

parameterised_impls = {
    ConcatenateOP: lambda op, *x: jnp.concatenate(x, axis=op.axis),
    ReshapeOP: lambda op, x: jnp.reshape(x, shape=op.newshape),
    NormOP: lambda op, x: jnp.linalg.norm(x, ord=op.ord),
}


def call_parameterised_op(op, *args):
    kls = op.__class__
    result = parameterised_impls[kls](op, *args)

    return result


def proj(i, n):
    p = np.zeros((n, n))
    p[i, i] = 1
    return p


def basis(i, n):
    p = np.zeros((n,))
    p[i] = 1
    return p


class JaxBackend(Backend):
    def __init__(self, *args, **kwargs):
        super(JaxBackend, self).__init__(*args, **kwargs)

    def native_types(self) -> Tuple[Type]:
        return (
            np.ndarray,
            np.int32,
            np.int64,
            np.float64,
            np.float32,
            float,
            complex,
            int,
        )

    def to_numpy_array(self, array) -> ArrayLike:
        if array is not None:
            return np.array(array)
        else:
            return array

    def to_backend_array(self, array):
        import scipy.sparse

        if scipy.sparse.issparse(array):
            array = array.toarray()
        return jnp.array(array)

    def reshape(self, arg, dim: Dimension):
        if dim.is_scalar():
            if isinstance(arg, scalar_types) or arg.ndim == 0:
                return arg
            else:
                try:
                    (inner,) = arg
                except ValueError as ex:
                    raise TypeError(f"Expecting a scalar, got {arg}") from ex
                return self.reshape(inner, dim)
        elif isinstance(arg, jnp.ndarray):
            return jnp.reshape(arg, dim.dim)
        elif isinstance(arg, np.ndarray):
            return np.reshape(arg, dim.dim)
        elif arg is None:
            return arg
        raise NotImplementedError(
            f"Don't know how to resize {arg.__class__.__name__}"
        )

    def call(self, op, *args) -> ArrayLike:

        try:
            result = impls[op](*args)
            return result
        except KeyError:
            pass

        if isinstance(op, tuple(parameterised_impls.keys())):
            return call_parameterised_op(op, *args)
        raise NotImplementedError(f"{op} is not implemented")

    def build_optimisation_problem(
        self,
        cost: Tracer,
        constraints: List[Tracer],
        arguments: List[Tracer],
        outputs: List[Tracer],
        **kwargs,
    ):
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Tuple, Dict
import numpy as np

from coker.toolkits.spatial import (
    Isometry3,
    Screw,
    SE3Adjoint,
    SE3CoAdjoint,
    hat,
    se3Adjoint,
)


@dataclass
class Inertia:
    centre_of_mass: Isometry3
    mass: float
    moments: np.ndarray

    @staticmethod
    def zero() -> "Inertia":
        return Inertia(Isometry3.identity(), 0, np.zeros((6,)))

    def as_matrix(self):
        inertial_component = np.array(
            [
                [self.moments[0], self.moments[1], self.moments[2]],
                [self.moments[1], self.moments[3], self.moments[4]],
                [self.moments[2], self.moments[4], self.moments[5]],
            ],
        )
        mass_component = self.mass * np.eye(
            3,
        )
        zeros = np.zeros(
            (3, 3),
        )
        return np.block([[inertial_component, zeros], [zeros, mass_component]])

    def as_matrix_origin(self):
        p_hat = hat(self.centre_of_mass.translation)
        assert self.centre_of_mass.rotation.angle == 0
        m = self.as_matrix()
        m[3:6, 0:3] = self.mass * p_hat
        m[0:3, 3:6] = -self.mass * p_hat
        return m


class JointType:
    @property
    def axes(self) -> List[Screw]:
        return []


class Weld(JointType):
    pass


class Free(JointType):

    @property
    def axes(self) -> List[Screw]:
        return [
            Screw.e_x(),
            Screw.e_y(),
            Screw.e_z(),
            Screw.w_x(),
            Screw.w_y(),
            Screw.w_z(),
        ]


class Revolute(JointType):
    def __init__(self, axis: Screw):
        self.axis = axis

    @property
    def axes(self) -> List[Screw]:
        return [self.axis]


class Planar(JointType):
    def __init__(self, *axes):
        self._axes = axes

    @property
    def axes(self) -> List[Screw]:
        return list(self._axes)


@dataclass
class RigidBody:
    WORLD = -1

    def __init__(self):
        """
        - g_i(q) is the position of the ith link
        - lam(i) is the parent of the ith link

        transforms:
            transform X_{lam(i), i} is the location of the origin of the ith
            link in lam(i) coordinates, so that at rest
            g_i(0) = prod_{j=0, i}(X_{lam(j), j} = X_{-1, 0} @ X_{0, 1} @ X_{lam(1), 1}) @ ...

        rest transforms:
            the rest transform is g_i(0) s.t. g_i(q) = exp(zeta q) g_i(0)


        """
        self.joints: List[JointType] = []

        self.inertia: List[Inertia] = []
        self.parents: List[int] = []
        self.transforms: List[Isometry3] = []
        self.end_effectors: List[Tuple[int, Isometry3]] = []
        self._rest_transforms: List[Isometry3] = []

    @property
    def joint_bases(self) -> List[List[Screw]]:
        return [j.axes for j in self.joints]

    def joint_global_basis(self, joint_index):
        bases = self.joint_bases[joint_index]

        adjoint = SE3Adjoint(self._rest_transforms[joint_index])
        return [adjoint.apply(b) for b in bases]

    def add_link(
        self, parent: int, at: Isometry3, joint: JointType, inertia: Inertia
    ) -> int:
        idx = len(self.parents)

        assert 0 <= parent < idx or parent == self.WORLD

        self.parents.append(parent)
        self.transforms.append(at)
        self.joints.append(joint)
        self.inertia.append(inertia)

        if parent != self.WORLD:
            t = self._rest_transforms[parent] @ at
        else:
            t = at
        self._rest_transforms.append(t)

        return idx

    def add_effector(self, parent: int, at: Isometry3):
        assert RigidBody.WORLD < parent < len(self.parents)
        idx = len(self.end_effectors)
        self.end_effectors.append((parent, at))
        return idx

    def total_joints(self):
        return sum([len(j) for j in self.joint_bases])

    def potential_energy(self, angles, gravity_vector):

        if len(angles.shape) != 1:
            assert angles.shape[0] == self.total_joints()
            q = np.reshape(angles, shape=(self.total_joints(),))
        else:
            assert angles.shape == (self.total_joints(),)
            q = angles
        joint_transforms = self._get_absolute_joint_xform(q)
        joint_xforms = self._accumulate_joint_xforms(joint_transforms)
        origin = np.array([0, 0, 0])
        total_energy = 0
        for i, inertia in enumerate(self.inertia):
            xform = (
                joint_xforms[i]
                @ self._rest_transforms[i]
                @ inertia.centre_of_mass
            )
            point = xform @ origin

            total_energy -= inertia.mass * np.dot(gravity_vector, point)

        return total_energy

    def _accumulate_joint_xforms(self, joint_transforms):

        accumulated_xform = []
        for i, parent in enumerate(self.parents):

            if parent != self.WORLD:
                t = accumulated_xform[parent] @ joint_transforms[i]
            else:
                t = joint_transforms[i]

            accumulated_xform.append(t)

        return accumulated_xform

    def _get_joint_local_bases(self):
        result = []
        for bases in self.joint_bases:
            for basis in bases:
                result.append(basis)
        return result

    def _get_joint_global_bases(self):
        result = []
        for rest_transform, bases in zip(
            self._rest_transforms, self.joint_bases
        ):
            for basis in bases:
                result.append(SE3Adjoint(rest_transform).apply(basis))
        return result

    def _get_joint_dependency_map(self):
        result = {}
        joint_idx = 0
        for i, bases in enumerate(self.joint_bases):
            result[i] = []
            for _ in bases:
                result[i].append(joint_idx)
                joint_idx += 1

        return result

    def get_dependant_joints(self, effector_index):
        joints = []

        parent, _ = self.end_effectors[effector_index]
        joint_map = self._get_joint_dependency_map()
        while parent != self.WORLD:
            joints += joint_map[parent]
            parent = self.parents[parent]
        return sorted(joints)

    def _get_joint_transforms(self, angles) -> List[Isometry3]:
        """Computes `g_si = Prod_j(exp(zeta_i q_i))`
        For each of the j angles.
        Zeta is the joint basis with respect to the joint origin.

        """

        joint_idx = 0
        transforms = []
        for link, bases in enumerate(self.joint_bases):
            g_theta = Isometry3.identity()
            for basis in bases:
                xform = basis.exp(angles[joint_idx])

                g_theta = g_theta @ xform
                joint_idx += 1
            transforms.append(g_theta)
        return transforms

    def _get_absolute_joint_xform(self, angles):
        r"""Express each g_j = Prod_i exp(theta_i zeta_i)
        in coordinates w.r.t the world

        The position of each joint $i$ is given by
        \prod_{j <= i} g_{j-1; j} exp(zeta_jq_j)

        where g_{i-1; j} is the transfrom of the joint origin to the parent frame
        zeta_j is the joint basis in the joint frame
        q_j is the joint angle

        want to move this to \Prod [exp(z_jq_j] g_i(0)

        # g_{j-1; j} = g_{j-1; 0} g_{0; j} = g_{j-1}^{-1} g_j
        so
            prod = g_0 exp(zeta_0) g^{-1}_0 * g_1 exp(zeta_1) g_1^{-1} ...
            i.e. z_j = Adj_{g_j}[zeta_j]

        """
        joint_idx = 0
        transforms = []
        for link, bases in enumerate(self.joint_bases):
            g_theta = Isometry3.identity()
            adjoint = SE3Adjoint(self._rest_transforms[link])
            for basis in bases:
                zeta = adjoint.apply(basis)
                g_theta = g_theta @ zeta.exp(angles[joint_idx])
                joint_idx += 1
            transforms.append(g_theta)
        return transforms

    def joint_transforms(self, angles, effector=None) -> List[Isometry3]:

        joint_index = 0
        transforms = []
        for parent, xform, bases in zip(
            self.parents, self.transforms, self.joint_bases
        ):
            g = transforms[parent] @ xform if parent != self.WORLD else xform
            for basis in bases:
                g = g @ basis.exp(angles[joint_index])
                joint_index += 1

            transforms.append(g)

        if effector is None:
            return transforms

        out = []
        parent, _ = self.end_effectors[effector]

        while parent != self.WORLD:
            out.append(transforms[parent])
            parent = self.parents[parent]
        return list(reversed(out))

    def forward_kinematics(self, angles) -> List[Isometry3]:
        xforms = self.joint_transforms(angles)
        return [xforms[parent] @ xform for parent, xform in self.end_effectors]

    def get_dependent_links(self, parent_link: int):
        result = list()
        while parent_link != self.WORLD:
            result.append(parent_link)
            parent_link = self.parents[parent_link]
        return reversed(result)

    def spatial_manipulator_jacobian(self, angles):
        return [
            self.spatial_single_manipulator_jacobian(angles, i)
            for i, _ in enumerate(self.end_effectors)
        ]

    def spatial_single_manipulator_jacobian(self, angles, end_effector):
        abs_xforms = self._get_absolute_joint_xform(angles)
        xforms = self._accumulate_joint_xforms(abs_xforms)

        effector_parent, _ = self.end_effectors[end_effector]
        dependent_link = set(self.get_dependent_links(effector_parent))

        columns = []

        for link_idx, bases in enumerate(self.joint_bases):
            if link_idx in dependent_link:
                parent = self.parents[link_idx]
                for zeta in bases:
                    zeta_prime = SE3Adjoint(
                        self._rest_transforms[link_idx]
                    ).apply(zeta)
                    if parent != self.WORLD:
                        zeta_prime = SE3Adjoint(xforms[parent]).apply(
                            zeta_prime
                        )
                    columns.append(
                        np.reshape(zeta_prime.to_array(), shape=(6, 1))
                    )
            else:
                zeta_prime = np.zeros((6, len(bases)))
                columns.append(zeta_prime)

        return np.concatenate(columns, axis=1)

    def body_manipulator_jacobian(self, angles):
        g_list = self.forward_kinematics(angles)
        ad_g_inv = [SE3Adjoint(g).inverse().as_matrix() for g in g_list]
        columns = [
            ad_g @ self.spatial_single_manipulator_jacobian(angles, i)
            for i, ad_g in enumerate(ad_g_inv)
        ]
        return columns

    def inverse_dynamics(
        self,
        angles,
        angle_rates,
        angle_accel,
        gravity: np.ndarray,
        end_effector_forces: Dict[int, Screw] = None,
    ) -> np.ndarray:
        """Compute torques for forces"""
        transforms = []
        velocities = []
        accelerations = []
        forces = []
        inverse_bases = []
        a_g = np.concatenate([np.zeros((3,)), gravity])

        x_forms = self._get_joint_transforms(angles)

        def spatial_cross(a, b):
            y_1 = np.cross(a[0:3], b[0:3])
            y_2 = np.cross(a[3:6], b[0:3]) + np.cross(a[0:3], b[3:6])
            return np.concatenate([y_1, y_2])

        def spatial_cross_star(a, b):
            y_1 = np.cross(a[0:3], b[0:3]) + np.cross(a[3:6], b[3:6])
            y_2 = np.cross(a[0:3], b[3:6])
            return np.concatenate([y_1, y_2])

        joint_idx = 0
        for link, bases in enumerate(self.joint_bases):
            parent = self.parents[link]
            s = []

            # g_i is the transform of the parent frame into the child frame
            # so that for a linear chain g_{01} @ g_{12} @ ... @ g_n
            # is the same as g_{0n}
            g_i = self.transforms[link] @ x_forms[link]

            vJ = np.zeros((6,))
            aJ = np.zeros((6,))
            for basis in bases:
                s_i = basis.to_array()
                s.append(s_i.reshape((6, 1)))
                vJ += s_i * angle_rates[joint_idx]
                aJ += s_i * angle_accel[joint_idx]
                joint_idx += 1

            adj = SE3Adjoint(g_i.inverse()).as_matrix()
            if parent != self.WORLD:
                v_i = adj @ velocities[parent] + vJ
                s_dot = spatial_cross(v_i, vJ)
                a_i = adj @ accelerations[parent] + aJ + s_dot
            else:
                v_i = vJ
                a_i = -adj @ a_g + aJ

            # Ineratia at center of mass,
            # need to transfrom to origin
            com_adj = SE3Adjoint(self.inertia[link].centre_of_mass.inverse())

            iota = (
                com_adj.as_matrix().transpose()
                @ self.inertia[link].as_matrix()
                @ com_adj.as_matrix()
            )
            f_i = iota @ a_i + spatial_cross_star(v_i, (iota @ v_i))
            transforms.append(g_i)
            velocities.append(v_i)
            accelerations.append(a_i)
            forces.append(f_i)
            if s:
                s_matrix = np.concatenate(s, axis=1)
                inverse_bases.append(s_matrix)
            else:
                inverse_bases.append(None)
        torques = []

        if end_effector_forces is not None:
            fk = self.forward_kinematics(angles)
            if isinstance(end_effector_forces, dict):
                iterator = end_effector_forces.items()
            else:
                iterator = enumerate(end_effector_forces)
            for idx, force in iterator:
                parent, _ = self.end_effectors[idx]
                forces[parent] -= SE3CoAdjoint(fk[idx]).as_matrix() @ force

        for i, S_i in reversed(list(enumerate(inverse_bases))):
            parent = self.parents[i]
            f_i = forces[i]
            if parent != self.WORLD:
                t = SE3CoAdjoint(transforms[i].inverse())
                parent_force = t.as_matrix() @ f_i
                forces[parent] = forces[parent] + parent_force

            if S_i is None:
                continue
            torque = S_i.T @ f_i
            torques = [torque] + torques
        result = np.concatenate(torques)
        return result

    def _get_link_com_jacobians(self, angles):
        joint_transforms = self._get_absolute_joint_xform(
            angles
        )  # exp(zeta_hat)
        joint_map = self._get_joint_dependency_map()  # link: [joint idx]
        joint_bases = self._get_joint_global_bases()  # zeta_j
        total_angles = self.total_joints()
        jacobians = []
        for i, inertia in enumerate(self.inertia):
            g = inertia.centre_of_mass @ self._rest_transforms[i]  # g_sli
            jacobian = [np.zeros((6, 1))] * total_angles
            next_idx = i

            while next_idx != self.WORLD:
                g = joint_transforms[next_idx] @ g
                for joint_idx in reversed(joint_map[next_idx]):
                    zeta_j = joint_bases[joint_idx]
                    adj_inv = SE3Adjoint(g).inverse()
                    zeta_j_dagger = adj_inv.apply(zeta_j)
                    jacobian[joint_idx] = np.reshape(
                        zeta_j_dagger.to_array(), shape=(6, 1)
                    )

                next_idx = self.parents[next_idx]

            jac = np.concatenate(jacobian, axis=1)
            jacobians.append(jac)

        return jacobians

    def mass_matrix(self, angles):
        jacobians = self._get_link_com_jacobians(angles)
        n = self.total_joints()
        mass_matrix = np.zeros((n, n))
        for j, inertia in zip(jacobians, self.inertia):

            mass_matrix += j.T @ inertia.as_matrix() @ j

        return mass_matrix

    def add_body(self, body: "RigidBody", at: Isometry3, parent: int):
        links = []

        iterator = zip(
            body.parents, body.joints, body.transforms, body.inertia
        )
        for i, (p_body, joint, xform, inertia) in enumerate(iterator):
            idx = len(self.parents)

            self.parents.append(
                parent if p_body == self.WORLD else links[p_body]
            )
            self.inertia.append(inertia)

            p_actual = self.parents[idx]
            if p_body == body.WORLD:
                body_xform = at @ xform

            else:
                body_xform = xform

            self.joints.append(joint)

            rest_xform = (
                self._rest_transforms[p_actual] @ body_xform
                if p_actual != self.WORLD
                else body_xform
            )
            self.transforms.append(body_xform)
            self._rest_transforms.append(rest_xform)

            links.append(idx)

        effectors = [
            self.add_effector(links[p_body], xform)
            for p_body, xform in body.end_effectors
        ]

        return links, effectors



"""Coker Backends

Supported
- Numpy/scipy (default)

To add:
- Jax
- Casadi
- Pytorch

"""

from coker.backends.evaluator import evaluate
from coker.backends.backend import get_backend_by_name, get_current_backend



from typing import Tuple, Type, Union
from coker import Dimension, Function, VectorSpace

from coker.backends.backend import Backend, ArrayLike

from coker.backends.casadi.casadi import *
from coker.backends.casadi.casadi import lower as _lower_to_casadi
from coker.backends.casadi.optimiser import build_optimisation_problem

from coker.backends.casadi.variational_solver import create_variational_solver
from coker.dynamics import VariationalProblem

scalar_types = (float, int)


class CasadiBackend(Backend):
    def to_numpy_array(self, array: Union[ca.MX, ca.DM]) -> ArrayLike:
        if isinstance(array, ca.MX):
            try:
                return array.to_DM().toarray()
            except RuntimeError:
                pass
        elif isinstance(array, ca.DM):
            return array.toarray()
        try:
            return ca.evalf(array).toarray()
        except RuntimeError:
            pass

        raise ValueError(f"Cannot convert {array} to a numpy array")

    def to_backend_array(self, array):
        import scipy.sparse

        if array is None:
            return ca.DM()
        if scipy.sparse.issparse(array):
            return ca.DM(scipy.sparse.csc_matrix(array))
        if isinstance(array, scalar_types):
            return ca.DM(array)
        if array.shape == (1, 1):
            return ca.DM(array[0, 0])
        elif array.shape == (1,):
            return ca.DM(array[0])
        elif len(array.shape) >= 2:
            result = ca.DM(array)

        elif len(array.shape) == 1 and array.shape[0] > 1:
            result = ca.DM(array.reshape(-1, 1))

        else:
            raise NotImplementedError(
                f"Don't know how to convert {array} to a casadi array"
            )

        assert not isinstance(result, Tracer)

        return result

    def call(self, op, *args) -> ArrayLike:
        try:
            result: ca.DM = impls[op](*args)
            assert result.is_regular(), f"{op}({args}) =  {result}"

            return result
        except KeyError:
            pass

        if isinstance(op, tuple(parameterised_impls.keys())):
            result = call_parameterised_op(op, *args)
            assert result.is_regular(), f"{op}({args}) =  {result}"
            return result

        if isinstance(op, ReshapeOP):

            (arg,) = args
            shape = op.newshape
            if len(shape) == 1:
                shape = (1, *shape)

            return ca.reshape(arg, shape)

        raise NotImplementedError(f"{op} is not implemented")

    def native_types(self) -> Tuple[Type]:
        pass

    def reshape(self, array: ArrayLike, dim: Dimension) -> ArrayLike:
        if dim.is_scalar():
            return array

        if dim.is_vector():
            shape = (*dim, 1)
        else:
            shape = tuple(dim)

        if isinstance(array, (ca.MX, ca.DM)):
            return ca.reshape(array, *shape)
        if isinstance(array, np.ndarray):
            return ca.reshape(array, *shape)
        raise NotImplementedError

    def _lower_with_evaluate(self, function):
        """Fallback: wraps CasadiBackend.evaluate for unsupported ca.Function cases."""
        backend = self

        def compiled(inputs):
            return backend.evaluate(function, inputs)

        return compiled

    def lower(self, function: Function):
        # Fall back to evaluate-based wrapper for cases that ca.Function cannot
        # handle: FunctionSpace inputs (partially evaluated functions) or None
        # outputs (e.g. from zero-row matmul).
        if any(
            isinstance(shape, FunctionSpace)
            for shape in function.input_shape()
        ):
            return self._lower_with_evaluate(function)

        ca_inputs, ca_outputs = _lower_to_casadi(
            function.tape, function.output
        )

        if any(o is None for o in ca_outputs):
            return self._lower_with_evaluate(function)
        ca_fn = ca.Function("f", ca_inputs, ca_outputs)
        backend = self
        fn_outputs = function.output

        def compiled(inputs):
            dm_inputs = [
                backend.to_backend_array(
                    np.asarray(a) if isinstance(a, list) else a
                )
                for a in inputs
                if a is not None
            ]
            dm_outs = ca_fn(*dm_inputs)
            if not isinstance(dm_outs, (list, tuple)):
                dm_outs = [dm_outs]

            result = []
            for y_i, output_tracer in zip(dm_outs, fn_outputs):
                try:
                    y_result = backend.to_numpy_array(y_i)
                    if output_tracer.dim.is_scalar():
                        if y_result.shape == (1, 1):
                            result.append(float(y_result[0, 0]))
                        elif y_result.shape == (1,):
                            result.append(float(y_result[0]))
                        else:
                            raise ValueError("Expected a scalar", y_result)
                    else:
                        result.append(y_result.reshape(output_tracer.shape))
                except ValueError:
                    result.append(y_i)
            return result

        return compiled

    def evaluate(self, function: Function, inputs: ArrayLike):
        workspace = {}

        for idx, (space, arg) in enumerate(
            zip(function.input_shape(), inputs)
        ):
            assert not isinstance(arg, Tracer)
            index = function.tape.input_indicies[idx]
            if isinstance(arg, np.ndarray):
                workspace[index] = self.to_backend_array(arg)

            else:
                workspace[index] = arg

        y = substitute(function.output, workspace)
        outs = []
        for y_i, output_tracer in zip(y, function.output):
            try:
                y_result = self.to_numpy_array(y_i)
                if output_tracer.dim.is_scalar():
                    if y_result.shape == (1, 1):
                        outs.append(float(y_result[0, 0]))
                    elif y_result.shape == (1,):
                        outs.append(float(y_result[0]))
                    else:
                        raise ValueError("Expected a scalar", y_result)
                else:
                    outs.append(y_result.reshape(output_tracer.shape))

            except ValueError:
                outs.append(y_i)

        return outs

    def to_array(self, arg: Union[ca.MX, ca.DM]):

        return self.to_numpy_array(arg)

    def build_optimisation_problem(
        self, cost, constraints, parameters, outputs, initial_conditions
    ):
        return build_optimisation_problem(
            cost, constraints, parameters, outputs, initial_conditions
        )

    def create_variational_solver(self, problem: VariationalProblem):
        return create_variational_solver(problem)

    def evaluate_integrals(
        self,
        functions,
        initial_conditions,
        end_point: float,
        inputs,
        solver_parameters=None,
    ):
        dxdt, g, dqdt = functions

        is_dae = dqdt is not Noop()
        has_quadrature = g is not Noop()
        x0, z0, q0 = (self.to_backend_array(a) for a in initial_conditions)
        if isinstance(end_point, (int, float)):
            if end_point == 0:
                return x0, z0, q0

            t_eval = [end_point]
        else:
            t_eval = end_point

        u, p = inputs
        p = self.to_backend_array(p)
        t = ca.MX.sym("t")
        x = ca.MX.sym("x", x0.shape)
        z = ca.MX.sym("z", z0.shape)

        dx_sym = dxdt(t, x, z, u, p)

        if is_dae:
            q = ca.MX.sym("q", q0.shape)
            q0 = ca.DM.zeros(q.shape)
            dq_sym = dqdt(t, x, z, u, p)
            txq = ca.vertcat(t, x, q)
            txq0 = ca.vertcat(ca.MX(0), x0, q0)
            xq_to_x_q = ca.Function("txq_to_x_q", [txq], [x, q])
            dtxq = ca.vertcat(ca.MX(1), dx_sym, dq_sym)
        else:
            txq = ca.vertcat(t, x)
            txq0 = ca.vertcat(ca.MX(0), x0)
            xq_to_x_q = ca.Function("txq_to_x_q", [txq], [x])
            dtxq = ca.vertcat(ca.MX(1), dx_sym)

        initial_conditions = {
            "x0": txq0,
        }
        dae = {
            "x": txq,
            "ode": dtxq,
        }
        if is_dae:
            dae["z"] = g(t, x, z, u, p)
            initial_conditions["z0"] = z0

        solver = ca.integrator("solver", "idas", dae, 0, t_eval, {})
        xq_final = solver(x0=txq0, z0=z0)

        if has_quadrature:
            x_final, q_final = xq_to_x_q(xq_final["xf"])
        else:
            x_final = xq_to_x_q(xq_final["xf"])
            q_final = None
        if is_dae:
            z_final = xq_final["zf"]
        else:
            z_final = None

        return x_final, z_final, q_final



import weakref

import numpy as np
import scipy as sp
from typing import Callable, Union, Tuple, List, Optional, Set, Iterable, Any
from collections import defaultdict


from coker.algebra.dimensions import (
    Dimension,
    VectorSpace,
    Scalar,
    FunctionSpace,
    Element,
)
from coker.algebra.tensor import SymbolicVector
from coker.algebra.ops import OP, Noop, Operator

from coker.algebra.ops import numpy_atomics, numpy_composites

import threading

scalar_types = (
    np.float32,
    np.float64,
    np.int32,
    np.int64,
    float,
    complex,
    int,
)


def get_basis(dimension: Dimension, i: int):
    return np.array([1 if j == i else 0 for j in range(dimension.dim[0])])


def get_projection(dimension: Dimension, slc: slice):
    if isinstance(dimension.dim, tuple):
        cols = dimension.dim[0]
    else:
        return 1

    indices = list(range(cols))[slc]
    rows = len(indices)
    proj = np.zeros((rows, cols), dtype=float)
    for row, col in enumerate(indices):
        proj[row, col] = 1
    return proj


Inferred = None


def get_dim_by_class(arg):
    if isinstance(arg, scalar_types):
        return Dimension(None)
    try:

        d = Dimension(arg.shape)
        return d
    except (AttributeError, ValueError):
        pass

    raise NotImplementedError(f"Don't know the shape of {type(arg)}")


class TapeInner:
    INNER_REF = -1
    CONSTANT_REF = -2

    def __init__(self, tape_ref: "Tape"):
        self._nodes = []
        self._constants = []
        self._constant_hashmap = {}
        self.tape_ref = weakref.ref(tape_ref)
        assert self.INNER_REF not in OP.__members__.values()
        assert self.CONSTANT_REF not in OP.__members__.values()

    @staticmethod
    def constant_hash(value) -> int:
        if isinstance(value, (int, float)):
            return hash((0, 0, value))
        elif isinstance(value, np.ndarray):
            return hash((*value.shape, *value.flatten().tolist()))
        else:
            return hash(value)

    @staticmethod
    def try_sparsify(value: Any) -> Any:
        if not isinstance(value, np.ndarray):
            return value

        n_entries: int = value.flatten().shape[0]

        if len(value.shape) == 1 or n_entries <= 16:
            return value

        nnz: int = np.count_nonzero(value)
        density = nnz / n_entries
        threshold = 0.25

        if density > threshold:
            return value
        try:
            return sp.sparse.csc_array(value)
        except ValueError:
            return value

    def push_op(self, op: OP, *args) -> int:
        assert isinstance(op, OP) or isinstance(op, Operator)
        idx = len(self._nodes)

        if op == OP.VALUE:
            (value,) = args

            hsh = self.constant_hash(value)
            if hsh in self._constant_hashmap:
                value_idx = self._constant_hashmap[hsh]
            else:
                value_idx = len(self._constants)
                self._constant_hashmap[hsh] = value_idx
                value = self.try_sparsify(value)
                self._constants.append(value)

            self._nodes.append((self.CONSTANT_REF, value_idx))
        else:
            self._nodes.append((op, *args))
        return idx

    def define_input(self, size: int) -> int:
        idx = len(self._nodes)
        assert isinstance(size, int)
        self._nodes.append((self.INNER_REF, idx, size))
        return idx

    def __getitem__(self, item):
        op, *args = self._nodes[item]
        if op == self.INNER_REF:
            idx, size = args
            return Tracer(self.tape_ref(), idx)
        if op == self.CONSTANT_REF:
            (value_idx,) = args
            return OP.VALUE, self._constants[value_idx]
        return op, *args

    def __len__(self):
        return len(self._nodes)


class Tape:
    NONE = -1
    MAP_TO_NONE = -2

    def __init__(self):
        self._inner = TapeInner(self)
        self.dim = []
        self.input_indicies = []
        self.input_names = []

    def op(self, i):
        return self.nodes[i][0]

    @property
    def nodes(self):
        return self._inner

    def find_dependents(self, tracer: "Tracer") -> Set[int]:
        if tracer is None or tracer is Noop():
            return set()

        index = tracer.index
        result = set()
        for i, inner in enumerate(self.nodes[tracer.index :]):
            if isinstance(inner, Tracer):
                continue
            op, *args = inner
            if op == OP.VALUE:
                continue
            for arg in args:
                if isinstance(arg, Tracer):
                    if arg.index == index or arg.index in result:
                        result.add(index + i)
        return result

    def inputs(self):
        for index in self.input_indicies:
            if index == Tape.NONE:
                yield None
            elif index == Tape.MAP_TO_NONE:
                yield Noop()
            else:
                yield Tracer(self, index)

    def __len__(self):
        return len(self.nodes)

    def __hash__(self):
        return id(self)

    def _compute_shape(self, op: OP, *args) -> Dimension:
        dims = []
        for arg in args:
            if arg is None:
                dims.append(None)
                continue

            assert isinstance(arg, Tracer)

            dims.append(arg.dim)

        return op.compute_shape(*dims)

    def append(self, op: OP, *args) -> int:
        args = [strip_symbols_from_array(a) for a in args]

        args = [
            self.insert_value(a) if not isinstance(a, Tracer) else a.copy()
            for a in args
        ]
        out_dim = self._compute_shape(op, *args)
        index = len(self.dim)
        self.nodes.push_op(op, *args)
        self.dim.append(out_dim)
        return index

    def insert_value(self, arg):
        if arg is None:
            return None
        assert not isinstance(arg, Tracer)
        dim = get_dim_by_class(arg)
        idx = len(self.dim)
        self.nodes.push_op(OP.VALUE, arg)
        self.dim.append(dim)
        return Tracer(self, idx)

    def input(self, v: VectorSpace | Scalar):
        if v is None:
            self.input_indicies.append(Tape.NONE)
            self.input_names.append("None")
            return None

        if isinstance(v, Noop):
            self.input_indicies.append(Tape.MAP_TO_NONE)
            self.input_names.append("Noop")
            return v

        if isinstance(v, VectorSpace):
            dimension = Dimension(v.dimension)
            self.dim.append(dimension)
            size = dimension.flat()
        elif isinstance(v, Scalar):
            self.dim.append(Dimension(None))
            size = 1
        elif isinstance(v, FunctionSpace):
            self.dim.append(v)
            size = sum([d.flat() for d in v.output_dimensions()])
        else:
            assert False, f"Invalid input type {v}: of {type(v)} "

        index = self.nodes.define_input(size)
        tracer = Tracer(self, index)

        self.input_indicies.append(index)
        self.input_names.append(v.name)
        return tracer

    def list_inputs(self) -> Iterable[None | Noop | Scalar | VectorSpace]:
        for arg_idx, node_idx in enumerate(self.input_indicies):
            if node_idx == Tape.NONE:
                yield None
            elif node_idx == Tape.MAP_TO_NONE:
                yield Noop()
            elif isinstance(self.dim[node_idx], FunctionSpace):
                yield self.dim[node_idx]
            else:
                dim: Dimension = self.dim[node_idx]
                name = self.input_names[arg_idx]
                yield dim.to_space(name)

    def substitute(self, index, value):
        assert index in self.input_indicies

        self.input_indicies.remove(index)
        self.nodes[index] = value


def is_additive_identity(space: Dimension, arg) -> bool:
    if isinstance(arg, scalar_types) and arg == 0:
        return True
    try:
        return (space.dim == arg.shape) and (arg == 0).all()
    except:
        pass

    return False


class Tracer(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(self, tape: Tape, index: int):
        self._tape = weakref.ref(tape)
        self.index = index

    @property
    def tape(self):
        return self._tape()

    def copy(self):
        return Tracer(self.tape, self.index)

    def is_input(self):
        return self.index in self.tape.input_indicies

    def is_constant(self):
        if self.is_input():
            return False
        op, *args = self.tape.nodes[self.index]
        if op != OP.VALUE:
            return False
        return True

    def is_functional(self):
        if self.is_input():
            return False
        op, *args = self.tape.nodes[self.index]
        if op not in {OP.EVALUATE}:
            return False
        return True

    def as_halfplane_bound(self) -> Tuple["Tracer", float, float]:
        op, lhs, rhs = self.tape.nodes[self.index]
        bounds = {
            OP.EQUAL: (-1e-9, 1e-9),
            OP.LESS_THAN: (0, np.inf),
            OP.LESS_EQUAL: (-1e-9, np.inf),
        }
        return rhs - lhs, *bounds[op]

    def value(self):
        op, *args = self.tape.nodes[self.index]

        if op != OP.VALUE:
            return self.tape.nodes[self.index]

        (arg,) = args
        return arg

    def __hash__(self):
        return hash(hash(self.tape) + self.index)

    def __repr__(self):
        return f"Tracer({self.index})"

    @property
    def shape(self) -> Tuple:
        dim = self.tape.dim[self.index]
        if dim.is_scalar():
            raise ValueError("Scalars have no shape")
        return dim.dim

    @property
    def dim(self) -> Dimension:
        return self.tape.dim[self.index]

    def __str__(self):
        return f"Tape {self.tape}:{self.index}"

    def __mul__(self, other):
        index = self.tape.append(OP.MUL, self, other)
        return Tracer(self.tape, index)

    def __rmul__(self, other):
        index = self.tape.append(OP.MUL, other, self)

        return Tracer(self.tape, index)

    def __add__(self, other):
        if is_additive_identity(other, self):
            return self

        if not isinstance(other, Tracer):
            other = self.tape.insert_value(other)

        if self.dim.is_scalar() and not other.dim.is_scalar():
            return self * np.ones(other.shape) + other
        elif not self.dim.is_scalar() and other.dim.is_scalar():
            return self + other * np.ones(self.shape)

        index = self.tape.append(OP.ADD, self, other)
        return Tracer(self.tape, index)

    def __radd__(self, other):
        if is_additive_identity(self.dim, other):
            return self

        if not isinstance(other, Tracer):
            other = self.tape.insert_value(other)

        return other + self

    def __sub__(self, other):
        index = self.tape.append(OP.SUB, self, other)
        return Tracer(self.tape, index)

    def __rmatmul__(self, other):
        index = self.tape.append(OP.MATMUL, other, self)

        assert other.shape[0] > 0
        return Tracer(self.tape, index)

    def __matmul__(self, other):
        index = self.tape.append(OP.MATMUL, self, other)
        return Tracer(self.tape, index)

    def __pow__(self, power, modulo=None):
        if isinstance(power, float) and power == 0.5:
            index = self.tape.append(OP.SQRT, self)
        elif isinstance(power, int):
            return self._do_integer_power(power)

        else:
            index = self.tape.append(OP.PWR, self, power)

        return Tracer(self.tape, index)

    @property
    def T(self):
        index = self.tape.append(OP.TRANSPOSE, self)
        return Tracer(self.tape, index)

    def _do_integer_power(self, power):
        if power <= 0:
            raise NotImplementedError("Negative power")
        result = self
        for _ in range(1, power):
            result = self * result

        return result

    def __getitem__(self, key):
        if self.is_constant():
            return self.value()[key]

        if isinstance(key, slice):
            dimension = self.tape.dim[self.index]
            if isinstance(dimension, FunctionSpace):
                assert len(dimension.output_dimensions()) == 1
                assert dimension.output_dimensions()[0].is_vector()
                p = get_projection(dimension.output_dimensions()[0], key)
            else:
                assert dimension.is_vector(), f"Tried to index a vector"
                p = get_projection(dimension, key)
            index = self.tape.append(OP.MATMUL, p, self)
            return Tracer(self.tape, index)

        elif isinstance(key, int):
            dimension = self.tape.dim[self.index]
            assert dimension.is_vector(), f"Tried to index a vector"
            p = get_basis(dimension, key)
            index = self.tape.append(OP.DOT, p, self)
            return Tracer(self.tape, index)

        raise NotImplementedError(
            "Cannot get key {}, not yet implemented", key
        )

    def __setitem__(self, key, value):
        if len(key) != len(self.shape):
            raise ValueError(
                f"Cannot set item {key} = {value} on {self} with shape {self.shape}"
            )

        # when we set an item, we need to do 2 things.
        # 1. Store the operation in the tape
        # 2. Mutate this object so that it points to the new tracer
        assert (
            self[key].shape == value.shape
        ), f"Expected shape {self[key].shape } but got {value.shape} for {key} = {value} on {self} with shape {self.shape}"

        # if the value here is a constant that is not referenced by any other
        # tracers, we can just go ahead and mutate it.
        op, old_value = self.tape.nodes[self.index]
        if (
            op == OP.VALUE
            and not self.tape.find_dependents(self)
            and (
                (isinstance(value, Tracer) and value.is_constant())
                or isinstance(value, scalar_types)
                or isinstance(value, np.ndarray)
            )
        ):
            old_value.__setitem__(key, value)
            return

        # otherwise, we need to add the "SET" operation to the tape
        # and change this tracers index to point to that.

        raise NotImplementedError(
            f"Cannot set item. SET operation not implemented yet for {key} = {value} on {self} with shape {self.shape}"
        )

    def __iter__(self):
        for i in range(self.shape[0]):
            yield self[i]

    def __le__(self, other):
        idx = self.tape.append(OP.LESS_EQUAL, self, other)
        return Tracer(self.tape, idx)

    def __ge__(self, other):
        idx = self.tape.append(OP.LESS_EQUAL, other, self)
        return Tracer(self.tape, idx)

    def __lt__(self, other):
        idx = self.tape.append(OP.LESS_THAN, self, other)
        return Tracer(self.tape, idx)

    def __gt__(self, other):
        idx = self.tape.append(OP.LESS_THAN, other, self)
        return Tracer(self.tape, idx)

    def __eq__(self, other):
        idx = self.tape.append(OP.EQUAL, self, other)
        return Tracer(self.tape, idx)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if (
            ufunc == np.matmul
            and isinstance(inputs[0], np.ndarray)
            and inputs[0].shape[0] == 0
        ):
            return Tracer(self.tape, self.tape.NONE)

        try:
            op = numpy_atomics[ufunc]

            index = self.tape.append(op, *inputs)
            return Tracer(self.tape, index)

        except KeyError:
            pass

        if ufunc == np.less:
            # lhs -> numpy item
            # rhs -> Tracer
            # op lhs < rhs
            lhs, rhs = inputs
            return rhs > lhs

        raise NotImplementedError(f"{ufunc} is not implemented")

    def __array_function__(self, func, types, args, kwargs):
        try:
            op = numpy_atomics[func]
            index = self.tape.append(op, *args)
            return Tracer(self.tape, index)
        except KeyError:
            pass

        try:
            op = numpy_composites[func](**kwargs)
            args = op.pre_process(*args)
            index = self.tape.append(op, *args)
            return Tracer(self.tape, index)
        except KeyError:
            pass

        raise NotImplementedError(f"{func} with {kwargs} is not implemented")

    def norm(self):
        assert (len(self.shape) == 1 and self.shape[0] > 1) or (
            len(self.shape) == 2 and self.shape[1] == 1
        )
        return np.sqrt(np.dot(self, self))

    def normalise(self):
        norm = self.norm()

        condition = norm == 0

        output = self.tape.append(OP.CASE, condition, self, self / norm)

        return Tracer(self.tape, output)

    def __call__(self, *args):
        index = self.tape.append(OP.EVALUATE, self, *args)
        return Tracer(self.tape, index)


class Function:
    INLINE_SIZE = 10

    def __init__(
        self, tape: Tape, outputs: List[Tracer], backend="coker", name=None
    ):
        self.name = name
        self.tape = tape
        self.backend = backend
        self._compiled = None
        if isinstance(outputs, Tracer) or outputs is None:
            self.output = [outputs]
            self.is_single = True
        else:
            self.output = outputs
            self.is_single = False

    @property
    def arguments(self):
        return self.tape.input_names.copy()

    def __repr__(self):
        name = self.name if self.name else "<unknown>"
        try:
            out_shape = f"{self.output_shape()}"
        except AttributeError:
            out_shape = f"{self.output}"
        return f"{name}:{self.input_shape()} -> {out_shape}"

    def to_space(self, name):
        return FunctionSpace(
            name,
            arguments=[
                dim.to_space("input_{i}") if dim else None
                for i, dim in enumerate(self.input_shape())
            ],
            output=[
                dim.to_space("output_{i}") if dim else None
                for i, dim in enumerate(self.input_shape())
            ],
        )

    def input_spaces(self):
        return list(self.tape.list_inputs())

    def input_shape(self) -> Tuple[Dimension, ...]:
        special_inputs = {
            Tape.NONE: None,
            Tape.MAP_TO_NONE: Noop().cast_to_function_space(None),
        }

        return tuple(
            self.tape.dim[i] if i >= 0 else special_inputs[i]
            for i in self.tape.input_indicies
        )

    def output_shape(self) -> Tuple[Dimension, ...]:
        return tuple(o.dim if o is not None else None for o in self.output)

    def _prepare_argument(self, arg, index):
        if index == Tape.MAP_TO_NONE:
            return Noop()
        elif index == Tape.NONE:
            return None

        elif isinstance(self.tape.dim[index], FunctionSpace):
            if isinstance(arg, Function):
                return arg

            return function(self.tape.dim[index].arguments, arg, self.backend)

        return arg

    def call_inline(self, *args) -> Tuple[Tracer]:
        from coker.backends import get_backend_by_name

        backend = get_backend_by_name("numpy", set_current=False)
        output = backend.evaluate(self, args)
        if self.is_single:
            return output[0]
        return output

    def __call__(self, *args):
        assert len(args) == len(
            self.tape.input_indicies
        ), f"Expected {len(self.tape.input_indicies)} arguments but got {len(args)}"

        args = [
            (self._prepare_argument(arg, idx))
            for idx, arg in zip(self.tape.input_indicies, args)
        ]

        from coker.backends import get_backend_by_name

        if any(isinstance(a, Tracer) for a in args):
            # Tracing context: interpret through numpy so ops are recorded on
            # the outer tape rather than evaluated numerically.
            backend = get_backend_by_name("numpy", set_current=False)
            output = backend.evaluate(self, args)
        else:
            # Concrete evaluation: compile on first call, reuse thereafter.
            if self._compiled is None:
                backend = get_backend_by_name(self.backend)
                self._compiled = backend.lower(self)
            output = self._compiled(args)

        if self.is_single:
            return output[0]
        return output

    def lower(self):
        from coker.backends import get_backend_by_name

        backend = get_backend_by_name(self.backend)
        return backend.lower(self)

    def compile(self, backend: str):
        raise NotImplementedError("Not yet implemented")

    def __le__(self, other: np.ndarray):
        # self < other
        assert len(self.output_shape()) == 1, "Cannot compare tensors"
        dim = self.output_shape()[0]
        assert dim.shape == get_dim_by_class(
            other
        ), "Arguments have different shapes"
        ones = np.ones_like(other)
        return InequalityExpression(self, -np.inf * ones, other, is_equal=True)

    def __ge__(self, other):
        # self => other
        assert len(self.output_shape()) == 1, "Cannot compare tensors"
        (dim,) = self.output_shape()
        assert dim == get_dim_by_class(
            other
        ), "Arguments have different shapes"
        ones = np.ones_like(other)
        return InequalityExpression(self, other, ones * np.inf, is_equal=True)

    def __lt__(self, other):
        # self <= other
        assert len(self.output_shape()) == 1, "Cannot compare tensors"
        (dim,) = self.output_shape()
        assert dim == get_dim_by_class(
            other
        ), "Arguments have different shapes"
        ones = np.ones_like(other)
        return InequalityExpression(
            self, -np.inf * ones, other, is_equal=False
        )

    def __gt__(self, other):
        # self > other
        assert len(self.output_shape()) == 1, "Cannot compare tensors"
        dim = self.output_shape()[0]
        assert dim.shape == other.shape, "Arguments have different shapes"
        ones = np.ones_like(other)
        return InequalityExpression(self, other, ones * np.inf, is_equal=False)


class InequalityExpression:
    def __init__(
        self,
        value: Function,
        lower: np.ndarray,
        upper: np.ndarray,
        is_equal: bool = False,
    ):
        self.value = value
        self.lower = lower
        self.upper = upper


def function(
    arguments: List[Scalar | VectorSpace | FunctionSpace],
    implementation: Callable[[Element, ...], Element],
    backend: str = "coker",
    name: Optional[str] = None,
) -> Function:
    # create symbols
    # call function to construct expression graph

    tape = Tape()
    with TraceContext() as tape:
        args = [tape.input(v) for v in arguments]
        result = implementation(*args)

        if isinstance(result, np.ndarray):
            result = strip_symbols_from_array(result)

        if isinstance(result, SymbolicVector):
            result = result.collapse()

        def wrap(v):
            if isinstance(v, np.ndarray):
                v = strip_symbols_from_array(v)
            if isinstance(v, SymbolicVector):
                v = v.collapse()

            if isinstance(v, Tracer):
                return v

            return tape.insert_value(v)

        if isinstance(result, (list, tuple)):
            result = [wrap(r) for r in result]

        else:
            result = wrap(result)

        if isinstance(result, Tracer) and result.index == tape.NONE:
            result = None

    return Function(tape, result, backend, name)


def strip_symbols_from_array(array: np.ndarray, float_type=float):
    if not isinstance(array, np.ndarray):
        return array

    symbols = defaultdict(list)

    with np.nditer(
        array, flags=["refs_ok", "multi_index"], op_flags=[["readwrite"]]
    ) as it:
        for x in it:
            try:
                x[...] = float_type(x)
            except TypeError as e:

                value = x.tolist()
                assert isinstance(
                    value, Tracer
                ), "Unexpected object in array: {}".format(value)
                symbols[value].append(it.multi_index)
                x[...] = 0.0

    symbol_array = array.astype(float)

    for symbol, coords in symbols.items():
        basis = np.zeros_like(array)
        for c in coords:
            basis[c] = 1
        symbol_array = symbol_array + basis * symbol

    return symbol_array


def normalise(v: Union[np.ndarray, Tracer]):
    if isinstance(v, np.ndarray):
        if all(v_i == 0 for v_i in v):
            return np.zeros_like(v), 0
        else:
            r = np.linalg.norm(v)
            return v / r, r

    assert isinstance(v, Tracer), f"Expected Tracer got {type(v)}"

    unit_v = v.normalise()
    norm_v = v.norm()

    return unit_v, norm_v


_comparison_ops = frozenset({OP.EQUAL, OP.LESS_THAN, OP.LESS_EQUAL})


def if_then_else(expression, true_branch, false_branch):
    if isinstance(expression, Tracer):
        node = expression.tape.nodes[expression.index]
        if isinstance(node, Tracer):
            # Raw input variable — not a comparison result.
            raise TypeError(
                "expression must result from a comparison operator (==, <, <=), "
                "got a raw input variable"
            )
        cond_op, *_ = node
        if cond_op not in _comparison_ops:
            raise TypeError(
                f"expression must result from a comparison operator (==, <, <=), got {cond_op}"
            )
        index = expression.tape.append(
            OP.CASE, expression, true_branch, false_branch
        )
        return Tracer(expression.tape, index)

    try:
        cond = bool(expression)
    except (ValueError, TypeError) as exc:
        raise TypeError(
            f"expression cannot be unambiguously coerced to bool: {exc}"
        ) from exc

    return true_branch if cond else false_branch


_local = threading.local()


class TraceContext:
    def __enter__(self):
        tape = Tape()
        _local.trace = tape
        return tape

    def __exit__(self, exc_type, exc_val, exc_tb):
        _local.trace = None

    @staticmethod
    def get_local_tape() -> Tape | None:
        return getattr(_local, "trace", None)



from typing import Tuple
from functools import reduce
from coker import Dimension
from coker.backends.coker.memory import MemorySpec
from coker.backends.coker.sparse_tensor import (
    dok_ndarray,
    scalar,
    tensor_vector_product,
    tensor_sum,
    cast_vector,
)
import numpy as np


def dense_array_cast(x):
    if isinstance(x, scalar):
        return np.array([x])
    return x


def _scale_tensor_by_array(
    tensor: dok_ndarray, scale: np.ndarray, n_output_dims: int
) -> dok_ndarray:
    """Return a copy of tensor where each entry is scaled by scale[output_indices].

    tensor has shape (*output_shape, *memory_shape).  The first n_output_dims
    index axes are the output axes; scale must have exactly that shape.
    """
    new_keys = {}
    for k, v in tensor.keys.items():
        factor = scale[k[:n_output_dims]]
        new_keys[k] = v * float(factor)
    return dok_ndarray(tensor.shape, new_keys)


class BilinearWeights(np.lib.mixins.NDArrayOperatorsMixin):

    def __init__(
        self,
        memory: MemorySpec,
        shape: Tuple[int, ...],
        constant=None,
        linear=None,
        quadratic=None,
    ):
        self.memory = memory

        assert isinstance(shape, tuple)
        self.shape = shape

        self.constant = dok_ndarray.from_maybe(constant, expected_shape=shape)
        self.linear = dok_ndarray.from_maybe(
            linear, expected_shape=(*shape, memory.count)
        )
        self.quadratic = dok_ndarray.from_maybe(
            quadratic, expected_shape=(*shape, memory.count, memory.count)
        )

    def transpose(self) -> "BilinearWeights":
        if len(self.shape) == 1:
            (n,) = self.shape
            # constant has shape (n,); build the (1, n) transpose explicitly
            transposed_constant = dok_ndarray(
                (1, n), {(0, k[0]): v for k, v in self.constant.keys.items()}
            )
            return BilinearWeights(
                self.memory,
                shape=(1, n),
                constant=transposed_constant,
                linear=self.linear.swap_indices(0, 1),
                quadratic=self.quadratic.swap_indices(0, 1),
            )
        if len(self.shape) == 2:
            n, m = self.shape
            return BilinearWeights(
                self.memory,
                shape=(m, n),
                constant=self.constant.T,
                linear=self.linear.swap_indices(0, 1),
                quadratic=self.quadratic.swap_indices(0, 1),
            )

        raise NotImplementedError(
            f"Cannot transpose {len(self.shape)} dimensions"
        )

    def __call__(self, x):
        x_v = dense_array_cast(x)
        try:
            qxx = (self.quadratic @ (x_v, x_v)).toarray()
        except TypeError as ex:
            raise ex

        ax = (self.linear @ x_v).toarray()

        c = self.constant.toarray()
        result = c + ax + qxx
        return np.reshape(result, self.shape)

    def diff(self, x):
        dq = tensor_vector_product(
            self.quadratic, x, axis=1
        ) + tensor_vector_product(self.quadratic, x, axis=2)

        return dq + self.linear

    def push_forwards(self, x, dx):
        x = dense_array_cast(x)
        dx = dense_array_cast(dx)

        dw = self.diff(x)
        qxx = self.quadratic @ (x, x)
        lx = self.linear @ x
        w = self.constant.clone() + lx + qxx
        return w.toarray(), (dw @ dx).toarray()

    def is_scalar(self):
        return self.shape == (1,)

    @property
    def is_constant(self):
        return not self.linear.keys and not self.quadratic.keys

    @property
    def is_linear(self):
        return not self.quadratic.keys

    @property
    def is_quadratic(self):
        return bool(self.quadratic.keys)

    def __mul__(self, other):
        if isinstance(other, scalar):
            constant = other * self.constant
            linear = other * self.linear
            quadratic = other * self.quadratic
            return BilinearWeights(
                self.memory,
                shape=self.shape,
                constant=constant,
                linear=linear,
                quadratic=quadratic,
            )

        try:
            assert all(s == 1 for s in other.shape) and not isinstance(
                other, BilinearWeights
            )
            return other.flat[0] * self
        except (AttributeError, AssertionError):
            pass

        if isinstance(other, np.ndarray) and other.shape == self.shape:
            # Componentwise scaling: result[i...] = self[i...] * other[i...]
            n_output_dims = len(self.shape)
            constant = _scale_tensor_by_array(
                self.constant, other, n_output_dims
            )
            linear = _scale_tensor_by_array(self.linear, other, n_output_dims)
            quadratic = _scale_tensor_by_array(
                self.quadratic, other, n_output_dims
            )
            return BilinearWeights(
                self.memory, self.shape, constant, linear, quadratic
            )

        if self.is_scalar():
            if isinstance(other, BilinearWeights):
                assert (
                    self.memory == other.memory
                ), f"Cannot multiply weights with different source"
                if self.is_constant:
                    return float(self.constant) * other
                if other.is_constant and other.is_scalar():
                    return float(other.constant) * self

                if self.is_linear and other.is_linear:
                    result = float(self.constant) * other
                    result.quadratic = dok_ndarray(
                        (1, 1, 1),
                        {(0, 0, 0): float(self.linear) * float(other.linear)},
                    )
                    return result

            if isinstance(other, (np.ndarray, dok_ndarray)):
                if isinstance(other, np.ndarray) and len(other.shape) == 1:
                    result_shape = other.shape
                else:
                    result_shape = None
                other = dok_ndarray.fromarray(other)
                if result_shape is None:
                    result_shape = other.shape
                constants = float(self.constant) * other

                # Other : (l, m)
                # self.linear: Array(1, n),         ->          (l, m, n)
                # self.quadratic : Array(1, n, n)   ->          (l, m, n, n)
                linear = outer_product(other, self.linear)
                quadratic = outer_product(other, self.quadratic)
                return BilinearWeights(
                    self.memory, result_shape, constants, linear, quadratic
                )

        raise TypeError(f"Cannot multiply {self} by {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __add__(self, other):
        if isinstance(other, scalar):
            linear = self.linear.clone()
            constant = self.constant + other
            quadratic = self.quadratic.clone()
            return BilinearWeights(
                self.memory,
                self.shape,
                constant=constant,
                linear=linear,
                quadratic=quadratic,
            )
        elif isinstance(other, BilinearWeights):

            if self.linear.is_empty() and self.quadratic.is_empty():
                return BilinearWeights(
                    other.memory,
                    self.shape,
                    self.constant + other.constant,
                    other.linear,
                    other.quadratic,
                )
            if other.linear.is_empty() and other.quadratic.is_empty():
                return BilinearWeights(
                    self.memory,
                    self.shape,
                    self.constant + other.constant,
                    self.linear,
                    self.quadratic,
                )

            assert self.memory == other.memory, f"{self}, {other}"
            return BilinearWeights(
                self.memory,
                self.shape,
                self.constant + other.constant,
                self.linear + other.linear,
                self.quadratic + other.quadratic,
            )
        elif isinstance(other, np.ndarray):
            return BilinearWeights(
                self.memory,
                self.shape,
                self.constant + dok_ndarray.fromarray(other),
                self.linear.clone(),
                self.quadratic.clone(),
            )

        raise TypeError(f"Cannot add {type(other)}")

    def __sub__(self, other):
        if isinstance(other, scalar):
            linear = self.linear.clone()
            constant = self.constant - other
            quadratic = self.quadratic.clone()
            return BilinearWeights(
                self.memory,
                self.shape,
                constant=constant,
                linear=linear,
                quadratic=quadratic,
            )
        elif isinstance(other, BilinearWeights):
            assert self.memory == other.memory
            return BilinearWeights(
                self.memory,
                self.shape,
                self.constant - other.constant,
                self.linear - other.linear,
                self.quadratic - other.quadratic,
            )
        raise TypeError(f"Cannot add {type(other)}")

    def __neg__(self):
        return BilinearWeights(
            self.memory,
            self.shape,
            -self.constant,
            -self.linear,
            -self.quadratic,
        )

    def __rmatmul__(self, other):

        if isinstance(other, (np.ndarray, dok_ndarray)):
            constant = other @ self.constant
            try:

                linear = other @ self.linear
            except IndexError as ex:
                print(f"{other}, {(self.linear.shape, self.linear.keys)}")
                raise ex
            quadratic = other @ self.quadratic
            shape = constant.shape
            return BilinearWeights(
                self.memory, shape, constant, linear, quadratic
            )

        raise TypeError(f"Cannot matmul {type(other)}")

    def __matmul__(self, other):
        assert (
            isinstance(other, BilinearWeights) and other.memory is self.memory
        )
        assert (
            (self.is_linear and other.is_linear)
            or self.is_constant
            or other.is_constant
        )
        # Contract self's last output axis with other's first output axis.
        # self.shape = (..., n), other.shape = (n, ...)
        # self.constant/linear/quadratic have shape (*self.shape, ...) — output axes first,
        # memory axes last.  We must contract over axis len(self.shape)-1 of self's tensors
        # and axis 0 of other's tensors, NOT the last (memory) axis.
        col = len(self.shape) - 1

        def _contract(lhs: dok_ndarray, rhs: dok_ndarray) -> dok_ndarray:
            """tensor_sum contracting lhs axis col with rhs axis 0."""
            return tensor_sum(lhs, rhs, l_index=col, r_index=0)

        constant = _contract(self.constant, other.constant)
        linear = _contract(self.constant, other.linear) + _contract(
            self.linear, other.constant
        )
        linear_linear_quadratic = tensor_sum(
            self.linear, other.linear, l_index=col, r_index=0
        )
        # tensor_sum produces (out_l..., mem_l, out_r..., mem_r).
        # BilinearWeights convention requires (out_l..., out_r..., mem_l, mem_r).
        # Move mem_l (at position col) past the M-1 out_r axes.
        mem_l_pos = col
        for _ in range(len(other.shape) - 1):
            linear_linear_quadratic = linear_linear_quadratic.swap_indices(
                mem_l_pos, mem_l_pos + 1
            )
            mem_l_pos += 1
        quadratic = (
            _contract(self.constant, other.quadratic)
            + _contract(self.quadratic, other.constant)
            + linear_linear_quadratic
        )
        shape = (*self.shape[:-1], *other.shape[1:])
        return BilinearWeights(self.memory, shape, constant, linear, quadratic)

    def compile(self) -> "_CompiledBW":
        """Pre-materialise dok_ndarray tensors into dense numpy arrays."""
        n = self.memory.count
        flat = int(np.prod(self.shape))
        c = self.constant.toarray().reshape(flat)
        L = (
            self.linear.toarray().reshape(flat, n)
            if self.linear.keys
            else None
        )
        Q = (
            self.quadratic.toarray().reshape(flat, n, n)
            if self.quadratic.keys
            else None
        )
        return _CompiledBW(self.memory, self.shape, c, L, Q)

    def clone(self):
        return BilinearWeights(
            self.memory,
            self.shape,
            self.constant.clone(),
            self.linear.clone(),
            self.quadratic.clone(),
        )

    def __array_ufunc__(self, ufunc, method, args, out=None):
        if ufunc == np.matmul and method == "__call__":
            return self.__rmatmul__(args)

        if ufunc == np.multiply and method == "__call__":
            return self.__mul__(args)

        if ufunc == np.add and method == "__call__":
            return self.__add__(args)

        if ufunc == np.subtract and method == "__call__":
            if self.is_scalar() and isinstance(args, scalar):
                if self.constant.is_empty():
                    constant = dok_ndarray((1,), {(0,): args})
                else:
                    constant = self.constant.clone()
                    constant[(0,)] = args - constant[(0,)]
                linear = -self.linear
                quadratic = -self.quadratic

                return BilinearWeights(
                    self.memory, self.shape, constant, linear, quadratic
                )

        raise NotImplementedError(f"{ufunc} not implemented")

    def __truediv__(self, other):
        if isinstance(other, scalar):
            return BilinearWeights(
                self.memory,
                self.shape,
                self.constant / other,
                self.linear / other,
                self.quadratic / other,
            )

        if isinstance(other, np.ndarray):
            if other.size == 1:
                return self / float(other)
            return self * (1.0 / other)

        raise TypeError(f"Cannot divide {self} by {type(other)}")

    def dot(self, rhs: "BilinearWeights"):
        """Matrix multiplication of two bilinear weights.

        We assume that the total order of the result is 2.
        So that either;
        - self and RHS have order <= 1 (i.e. the quadratic terms are zero)
        - or one has order 2 and the other has order 0.

        If
        math::
            y_0 = c_0 + L_0x + Q_0(x,x)
            y_1 = c_1 + L_1x + Q_1(x,x)

            dot(y_0, y_1) = dot(c_0, c_1)
                             + ((c_0.T @ L_1 + c_1.T @ L_0 x)
                             + (c_0.T @ Q_1 + c_1.T @ Q_0 + L_0.T @ L_1) (x,x)


        """
        assert self.memory == rhs.memory
        assert (
            (self.is_linear and rhs.is_linear)
            or (not self.is_linear and rhs.is_constant)
            or (self.is_constant and not rhs.is_linear)
        ), "dot requires both operands order<=1, or one order==2 and the other order==0"
        memory_count = self.memory.count
        output_size = int(np.prod(self.shape))
        c_self = self.constant.toarray().flatten()  # (output_size,)
        c_rhs = rhs.constant.toarray().flatten()  # (output_size,)
        l_self = self.linear.toarray().reshape(output_size, memory_count)
        l_rhs = rhs.linear.toarray().reshape(output_size, memory_count)
        q_self = self.quadratic.toarray().reshape(
            output_size, memory_count, memory_count
        )
        q_rhs = rhs.quadratic.toarray().reshape(
            output_size, memory_count, memory_count
        )

        constant_val = float(np.dot(c_self, c_rhs))
        c = dok_ndarray.fromarray(np.array([constant_val]))

        linear_val = (c_self @ l_rhs + c_rhs @ l_self).reshape(1, memory_count)
        l = dok_ndarray.fromarray(linear_val)

        linear_linear = (l_self.T @ l_rhs).reshape(
            1, memory_count, memory_count
        )
        c_self_quadratic = np.einsum("i,ijk->jk", c_self, q_rhs).reshape(
            1, memory_count, memory_count
        )
        c_rhs_quadratic = np.einsum("i,ijk->jk", c_rhs, q_self).reshape(
            1, memory_count, memory_count
        )
        q = dok_ndarray.fromarray(
            linear_linear + c_self_quadratic + c_rhs_quadratic
        )

        return BilinearWeights(self.memory, (1,), c, l, q)

    @staticmethod
    def identity2(memory: MemorySpec):

        shape = (memory.count,)
        data = {}
        for i in range(memory.count):
            data[(i, i)] = 1

        linear = dok_ndarray((memory.count, memory.count), data)

        return BilinearWeights(memory, shape, linear=linear)

    @staticmethod
    def reshape_identity(memory: MemorySpec, shape: tuple):
        """BilinearWeights that maps a flat vector of size memory.count to the given shape."""
        n = memory.count
        data = {}
        for k in range(n):
            multi_idx = np.unravel_index(k, shape, order="C")
            data[(*multi_idx, k)] = 1
        linear = dok_ndarray((*shape, n), data)
        return BilinearWeights(memory, shape, linear=linear)


class _CompiledBW:
    """Dense pre-materialised BilinearWeights for fast runtime evaluation.

    Created by BilinearWeights.compile(). Stores constant/linear/quadratic
    as plain numpy arrays so __call__ is a pure BLAS operation with no
    Python dict iteration.
    """

    def __init__(self, memory, shape, c, L, Q):
        self.memory = memory
        self.shape = shape
        self._c = c  # numpy (flat_out,)
        self._L = L  # numpy (flat_out, n_mem) or None
        self._Q = Q  # numpy (flat_out, n_mem, n_mem) or None

    def __call__(self, x):
        result = self._c.copy()
        if self._L is not None:
            result = result + self._L @ x
        if self._Q is not None:
            result = result + np.einsum("ijk,j,k->i", self._Q, x, x)
        return result.reshape(self.shape)

    def push_forwards(self, x, dx):
        result = self._c.copy()
        jac = np.zeros((result.size, len(x)))
        if self._L is not None:
            result = result + self._L @ x
            jac = jac + self._L
        if self._Q is not None:
            result = result + np.einsum("ijk,j,k->i", self._Q, x, x)
            jac = jac + np.einsum(
                "ijk,k->ij", self._Q + self._Q.transpose(0, 2, 1), x
            )
        return result.reshape(self.shape), (jac @ dx).reshape(self.shape)


def outer_product(lhs: dok_ndarray, rhs: dok_ndarray):
    assert rhs.shape[0] == 1
    shape = (*lhs.shape, *rhs.shape[1:])
    data = {}
    for k_l, v_l in lhs.keys.items():
        for k_r, v_r in rhs.keys.items():
            key = tuple((*k_l, *k_r[1:]))
            data[key] = v_l * v_r

    return dok_ndarray(shape, data)



import numpy as np
import xml.etree.ElementTree as ET

from coker.toolkits.kinematics import RigidBody, Inertia
from coker.toolkits.spatial import Rotation3


def to_string(item) -> str:
    if isinstance(item, np.ndarray):
        return " ".join([f"{x}" for x in item.tolist()])
    if isinstance(item, Rotation3):
        return " ".join([f"{x}" for x in item.as_quaternion().to_elements()])

    if isinstance(item, list):
        return " ".join([f"{i}" for i in item])

    raise NotImplementedError


def build_inertial_tags(inertial: Inertia):
    pos = to_string(inertial.centre_of_mass.translation)
    fullinertia = to_string(
        [
            inertial.moments[0],
            inertial.moments[3],
            inertial.moments[5],
            inertial.moments[1],
            inertial.moments[2],
            inertial.moments[4],
        ]
    )

    return dict(pos=pos, fullinertia=fullinertia, mass=f"{inertial.mass}")


def to_mujoco_xml(body: RigidBody, name: str) -> ET.ElementTree:
    root = ET.Element("mujoco")
    root.attrib["model"] = name
    world = ET.SubElement(root, "worldbody")
    floor = ET.SubElement(world, "geom")
    floor.attrib.update(
        dict(name="floor", size="0 0 0.05", type="plane", condim="3")
    )

    bodies = []
    joints = []
    for i, (parent_idx, xform, joint, inertia) in enumerate(
        zip(body.parents, body.transforms, body.joints, body.inertia)
    ):

        if parent_idx == body.WORLD:
            this_body = ET.SubElement(world, "body")
        else:
            this_body = ET.SubElement(bodies[parent_idx], "body")

        pos = " ".join([f"{x}" for x in xform.translation.tolist()])

        this_body.attrib.update(dict(name=f"body_{i}", pos=pos))

        inertia_el = ET.SubElement(this_body, "inertial")
        inertia_el.attrib.update(build_inertial_tags(inertia))

        bodies.append(this_body)

        if xform.rotation is not None and xform.rotation != Rotation3.zero():
            orientation = " ".join(
                [f"{q}" for q in xform.rotation.as_quaternion().to_elements()]
            )
            this_body.attrib.update(dict(quat=orientation))

        if len(joint.axes) == 6:
            ET.SubElement(this_body, "freejoint")
            continue

        for axis in joint.axes:
            j = len(joints)
            joint_el = ET.SubElement(this_body, "joint")
            joints.append(joint_el)
            if (axis.translation == np.array([0, 0, 0])).all():
                rot = " ".join([f"{x}" for x in axis.rotation])
                joint_el.attrib.update(dict(name=f"q_{j}", axis=rot))
            else:
                raise NotImplementedError

        parent_bdy = bodies[parent_idx]
        geom = ET.SubElement(parent_bdy, "geom")

        geom.attrib.update(
            dict(
                type="capsule",
                size="0.01",
                fromto="0 0 0 " + to_string(xform.apply(np.zeros((3,)))),
            )
        )

    for parent_idx, xform in body.end_effectors:
        parent = bodies[parent_idx]
        geom = ET.SubElement(parent, "geom")

        geom.attrib.update(
            dict(
                type="capsule",
                size="0.01",
                fromto="0 0 0 " + to_string(xform.translation),
            )
        )
    return ET.ElementTree(root)



from abc import ABCMeta, abstractmethod
from typing import Any, Tuple, Type
from coker import Function
from coker import Tracer
from typing import List, Dict

from coker.dynamics import (
    VariationalProblem,
    create_autonomous_ode,
    DynamicsSpec,
)

ArrayLike = Any


class SolverParameters(metaclass=ABCMeta):
    """Interface for backend-specific ODE solver configuration."""

    pass


class Backend(metaclass=ABCMeta):

    @abstractmethod
    def native_types(self) -> Tuple[Type]:
        pass

    @abstractmethod
    def to_numpy_array(self, array) -> ArrayLike:
        """Cast array from backend to numpy type."""
        pass

    @abstractmethod
    def to_backend_array(self, array: ArrayLike):
        """Cast array from native python (numpy) to backend type."""
        pass

    @abstractmethod
    def reshape(self, array: ArrayLike, shape: Tuple[int, ...]) -> ArrayLike:
        pass

    @abstractmethod
    def call(self, op, *args) -> ArrayLike:
        pass

    @abstractmethod
    def build_optimisation_problem(
        self,
        cost: Tracer,  # cost
        constraints: List[Tracer],
        parameters: List[Tracer],
        outputs: List[Tracer],
        initial_conditions: Dict[int, ArrayLike],
    ):
        raise NotImplementedError

    def create_variational_solver(self, problem: VariationalProblem):
        raise NotImplementedError

    def resolve_fn(self, op):
        """Return a callable for op, resolved at plan-build time.

        The default wraps backend.call; backends can override to return the
        underlying function directly and avoid the per-call dispatch overhead.
        """
        _op = op
        return lambda *args: self.call(_op, *args)

    def resolve_post_fn(self, dim):
        """Return the post-processing function applied to a step's output.

        The default handles the general case (Tracer passthrough + reshape).
        Backends can override to return an identity function where safe, removing
        the isinstance check and reshape call from the hot loop.
        """
        _dim = dim
        _reshape = self.reshape

        def post(value):
            if not isinstance(value, Tracer):
                return _reshape(value, _dim)
            return value

        return post

    def evaluate(self, function: Function, inputs: ArrayLike):
        from coker.backends.evaluator import evaluate_inner

        workspace = {}
        return evaluate_inner(
            function.tape, inputs, function.output, self, workspace
        )

    def evaluate_integrals(
        self,
        functions,
        initial_conditions,
        end_point: float,
        inputs,
        solver_parameters=None,
    ):
        raise NotImplementedError(
            "Evaluating integrals is not implemented for this backend"
        )

    def lower(self, function: Function):
        """Compile function to a callable for repeated numerical evaluation.

        Called once on the first concrete (non-Tracer) invocation of
        Function.__call__. Returns a callable f(inputs) -> outputs where
        inputs and outputs are lists matching the function's signature.

        The default wraps evaluate_inner; backends override this to return an
        optimised compiled callable (e.g. a plan-based closure for numpy, a
        ca.Function for casadi).
        """
        backend = self
        tape = function.tape
        outputs = function.output

        def compiled(inputs):
            from coker.backends.evaluator import evaluate_inner

            workspace = {}
            return evaluate_inner(tape, inputs, outputs, backend, workspace)

        return compiled


__known_backends = {}


def instantiate_backend(name: str):
    if name == "numpy":
        import coker.backends.numpy.core

        backend = coker.backends.numpy.core.NumpyBackend()
    elif name == "coker":
        import coker.backends.coker

        backend = coker.backends.coker.CokerBackend()
    elif name == "jax":
        import coker.backends.jax

        backend = coker.backends.jax.JaxBackend()
    elif name == "casadi":
        import coker.backends.casadi

        backend = coker.backends.casadi.CasadiBackend()
    elif name == "sympy":
        import coker.backends.sympy

        backend = coker.backends.sympy.SympyBackend()
    else:
        raise ValueError(f"Unknown backend: {name}")

    __backends[name] = backend
    return backend


def get_backend_by_name(name: str, set_current=True) -> Backend:
    try:
        b = __backends[name]
        if set_current:
            __current_backend = b
        return b
    except KeyError:
        pass
    try:
        b = instantiate_backend(name)
        if set_current:
            __current_backend = b
        return b
    except KeyError:
        pass

    raise NotImplementedError(f"Unknown backend {name}")


__backends = {}

default_backend = "coker"
__current_backend = None


def get_current_backend() -> Backend:
    if __current_backend is None:
        return get_backend_by_name(default_backend)
    else:
        return __current_backend



from coker.algebra.kernel import (
    Tape,
    Tracer,
    Function,
    function,
    scalar_types,
    normalise,
    get_projection,
    if_then_else,
)
from coker.algebra.dimensions import (
    Dimension,
    VectorSpace,
    Scalar,
    FunctionSpace,
    Element,
)
from coker.algebra.ops import OP, Noop
from coker.algebra.tensor import SymbolicVector

from coker.algebra.factories import zeros



from typing import Union
import numpy as np


def is_scalar(x: Union[int, float, np.ndarray]):
    if isinstance(x, np.ndarray):
        return all(s_i == 1 for s_i in x.shape)
    return isinstance(x, (int, float))



from coker.backends.casadi.casadi import substitute, to_casadi, lower
from typing import List
from coker import Tracer

import casadi as ca
import numpy as np


def build_optimisation_problem(
    cost, constraints, parameters: List[Tracer], outputs, initial_conditions
):

    # p = P(parameters)
    # x = P(inputs ~ parameter)
    tape = cost.tape
    workspace = {}
    parameter_indicies = {p.index for p in parameters}
    inputs = [i for i in tape.input_indicies if i not in parameter_indicies]

    parameter_dim = sum(p.dim.flat() for p in parameters)
    input_dim = sum(tape.dim[i].flat() for i in inputs)

    x = ca.MX.sym("x", input_dim)
    x0 = ca.DM(input_dim, 1)
    input_offset = 0
    parameter_offset = 0

    for i in inputs:
        n_i = tape.dim[i].flat()
        projection = ca.MX(n_i, input_dim)
        dm_projection = ca.DM(n_i, input_dim)
        for j in range(n_i):
            projection[j, j + input_offset] = 1
            dm_projection[j, j + input_offset] = 1

        workspace[i] = projection @ x
        x0_i = initial_conditions[i]
        x0 += dm_projection.T @ ca.DM(x0_i)

        input_offset += n_i

    p = ca.MX.sym("p", parameter_dim)

    for i in parameter_indicies:
        n_i = tape.dim[i].flat()
        projection = ca.MX(n_i, parameter_dim)
        for j in range(n_i):
            projection[j, parameter_offset + j] = 1

        workspace[i] = projection @ p
        parameter_offset += n_i

    (cost_fn,) = substitute([cost], workspace)
    output_map = ca.Function("y", *lower(tape, outputs, workspace))

    n_constraints = len(constraints)

    #    constraint_function = np.zeros((n_constraints,))
    #    lower_bound = ca.DM(n_constraints, 1)
    #    upper_bound = ca.DM(n_constraints, 1)
    cs = []
    lbs = []
    ubs = []

    for i, constraint in enumerate(constraints):
        c, lb, ub = constraint.as_halfplane_bound()
        (c_i,) = substitute([c], workspace)

        lbs.append(to_casadi(lb) * ca.DM.ones(*c_i.shape))
        ubs.append(to_casadi(ub) * ca.DM.ones(*c_i.shape))
        cs.append(c_i)

    upper_bound = ca.vertcat(*ubs)
    lower_bound = ca.vertcat(*lbs)
    g = ca.vertcat(*cs)

    spec = {"x": x, "p": p, "f": cost_fn, "g": g}

    solver_inner = ca.nlpsol("solver", "ipopt", spec)

    return CasadiSolver(
        solver_inner, p, None, (lower_bound, upper_bound), output_map, x0
    )


class CasadiSolver:
    def __init__(self, solver_inner, p, x_bounds, g_bounds, output_map, x0):
        self.solver_inner = solver_inner
        self.x_bounds = x_bounds
        self.g_bounds = g_bounds
        self.output_map = output_map
        self.x0 = x0
        self.p: ca.MX = p

    def __call__(self, *args):
        if not self.p.is_empty():
            raise NotImplementedError

        spec = {
            "x0": self.x0,
            "lbg": self.g_bounds[0],
            "ubg": self.g_bounds[1],
        }
        soln = self.solver_inner(**spec)

        result = self.output_map(soln["x"])

        return [r.full() for r in result]



from collections import defaultdict
from typing import Set, Dict, Tuple

from coker import Function, Tracer
import numpy as np
from coker.backends.coker.layers import InputLayer, OutputLayer
from coker.backends.coker.memory import MemorySpec
from coker.backends.coker.op_impl import *


def label_sinks(function: Function) -> Tuple[Set[int], Set[int]]:
    """
    Summary:

        Forward pass through the graph, determining which nodes are
        to be considered "sources"

        Criteria is either
            a) nonlinear
            b) used as inputs to multiple different nonlinear terms

    """
    tape = function.tape
    constants = set()
    tape_outdegree = [0] * len(tape)
    sources = {}
    sink_nodes = {o.index for o in function.output if o is not None}
    # output of these nodes are \considered 'new variables'

    for i, node in enumerate(tape.nodes):

        if i in tape.input_indicies:
            sources[i] = [-1]
            sink_nodes.add(i)
            continue
        else:
            sources[i] = []

        op, *args = node
        if op == OP.VALUE:
            constants.add(i)
            continue

        indices = [a.index for a in args]
        in_nodes = [idx for idx in indices if idx not in constants]

        if not in_nodes:
            constants.add(i)
            continue

        # non-constant op
        #

        # Strictly Linear nodes
        if op.is_linear():
            for j in in_nodes:
                for source in sources[j]:
                    if source not in sources[i]:
                        sources[i].append(source)
            continue

        for j in in_nodes:
            sources[i] += sources[j]

        # Multi-linear terms that mayne nonlinear
        if op.is_bilinear():
            if len(set(sources[i])) == 1:
                continue

        if op == OP.DIV and indices[1] in constants:
            continue

        sink_nodes.add(i)

    for i, degree in enumerate(tape_outdegree):
        if degree >= 2:
            sink_nodes.add(i)

    return sink_nodes, constants


def label_layers(function: Function, sink_nodes: Dict):
    edges = defaultdict(set)
    tape = function.tape
    distance = [0] * len(tape)

    def recurse_node(sink, node, depth):
        if node in tape.input_indicies:
            edges[sink].add(node)
            return
        op, *args = tape.nodes[node]

        if node is tape.NONE or node is tape.MAP_TO_NONE:
            edges[sink].add(node)
            return

        assert not isinstance(op, Tracer)
        if op == OP.VALUE:
            edges[node].add(sink)
            return

        for a in args:
            idx = a.index
            edges[node] |= {sink}
            if idx in sink_nodes:
                distance[idx] = max(distance[idx], depth + 1)
                edges[sink].add(a.index)
            else:
                recurse_node(sink, a.index, depth)

    for o in sorted(sink_nodes, reverse=True):
        recurse_node(o, o, distance[o])

    edges.update({i: {i} for i in tape.input_indicies})
    max_layers = max(distance)
    distance = [max_layers - d for d in distance]

    return edges, distance


def label_sources(
    function: Function, sink_nodes=None, constants=None
) -> Dict[int, Set[int]]:
    """

    Starting with the inputs and sink nodes, label all downstream nodes that depend on those sinks

    we then end up with

    """
    if sink_nodes is None or constants is None:
        sink_nodes, constants = label_sinks(function)

    arguments = {i: set() for i in constants}
    arguments.update({i: {i} for i in sink_nodes})
    arguments.update({i: {i} for i in function.tape.input_indicies})
    workset = [i for i in range(len(function.tape)) if i not in arguments]

    for idx in workset:
        op, *args = function.tape.nodes[idx]

        arguments[idx] = set.union(*(arguments[a.index] for a in args))

    return arguments


class SparseNet:
    def __init__(
        self,
        memory,
        input_layer: InputLayer,
        output_layer: OutputLayer,
        intermediate_layers,
    ):
        self.memory = memory
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.intermediate_layers = intermediate_layers

    @property
    def layers(self):
        return [self.input_layer, *self.intermediate_layers, self.output_layer]

    def __call__(self, *args):
        workspace = self.apply_input_map(*args)

        for layer in self.intermediate_layers:
            in_specs = layer.inputs()
            (out_spec,) = layer.outputs()
            out = layer(*[workspace[k] for k in in_specs])
            workspace[out_spec] = np.asarray(out).flatten()
        return self.output_layer.call(workspace)

    def push_forward(self, *tangent_spaces):
        n_args = len(self.input_layer.vec_to_arg_maps)
        x, dx = tangent_spaces[0:n_args], tangent_spaces[n_args:]
        workspace = self.apply_input_map(*x)
        dworkspace = self.apply_input_map(*dx)

        for layer in self.intermediate_layers:
            in_specs = layer.inputs()
            (out_spec,) = layer.outputs()
            x_i = [workspace[k] for k in in_specs]
            dx_i = [dworkspace[k] for k in in_specs]
            out, dout = layer.push_forward(*x_i, *dx_i)
            workspace[out_spec] = out
            dworkspace[out_spec] = dout

        y = self.output_layer.call(workspace)
        dy = self.output_layer.call(dworkspace)
        return y, dy

    def apply_input_map(self, *args) -> Dict[MemorySpec, np.ndarray]:
        return {self.memory[0]: self.input_layer(*args)}

    def apply_output_map(self, context):
        return self.output_layer.call(context)



import enum
import numpy as np
from typing import Dict, Callable
from coker.algebra.exceptions import InvalidShape, InvalidArgument
from coker.algebra.dimensions import Dimension, FunctionSpace
from typing_extensions import final


class OP(enum.Enum):
    VALUE = 0
    ADD = 1
    SUB = 2
    MUL = 3
    DIV = 4
    MATMUL = 5
    INT_PWR = 6
    PWR = 7
    EXP = 8
    SIN = 9
    COS = 10
    TAN = 11
    ARCSIN = 12
    ARCCOS = 13
    ARCTAN = 14
    SQRT = 15
    DOT = 16
    CROSS = 17
    TRANSPOSE = 18
    NEG = 19
    ABS = 20
    CASE = 21
    EQUAL = 22
    ARCTAN2 = 23
    LESS_THAN = 24
    LESS_EQUAL = 25
    EVALUATE = 26
    LOG = 27

    def compute_shape(self, *dims: Dimension) -> Dimension:
        return compute_shape[self](*dims)

    def is_linear(self):
        return self in {OP.ADD, OP.SUB, OP.TRANSPOSE, OP.NEG}

    def is_bilinear(self):
        return self in {OP.MUL, OP.MATMUL, OP.CROSS, OP.DOT}

    def is_nonlinear(self):
        return not self.is_linear() and not self.is_bilinear()


@final
class Noop:
    def __call__(self, *args, **kwargs):
        return None

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance

    @staticmethod
    def cast_to_function_space(arguments=None):
        return FunctionSpace("noop", arguments, None)


class Operator:
    def pre_process(self, *args):
        return args

    def compute_shape(self, *dims: Dimension) -> Dimension:
        raise NotImplementedError("Compute shape not implemented for {}", self)

    def is_linear(self):
        raise NotImplementedError("is_linear not implemented for {}", self)

    def is_bilinear(self):
        return False

    def is_nonlinear(self):
        return not self.is_linear() and not self.is_bilinear()


class ConcatenateOP(Operator):
    __slots__ = ("axis",)

    def __init__(self, axis=0):
        self.axis = axis

    def is_linear(self):
        return True

    def pre_process(self, *args):
        # Args should be a single argument containing a list of things to concatenate
        assert len(args) == 1
        return args[0]

    def compute_shape(self, *dims: Dimension) -> Dimension:

        if all(d.is_scalar() or d.is_vector() for d in dims):
            dim = Dimension((sum(d.flat() for d in dims),))
            return dim

        out_dims = list(dims[0].dim)
        for d in dims[1:]:
            assert all(
                d.dim[i] == out_dims[i]
                for i in range(len(out_dims))
                if i != self.axis
            )
            out_dims[self.axis] += d.dim[self.axis]

        return Dimension(tuple(out_dims))


class ReshapeOP(Operator):
    def __init__(self, shape, order="C"):
        self.newshape = shape
        self.order = order

    def compute_shape(self, dim: Dimension) -> Dimension:
        return Dimension(self.newshape)

    def is_linear(self):
        return True

    def pre_process(self, value, *args):
        if self.order == "C":
            return value, *args
        else:
            return value.T, *args


class NormOP(Operator):
    def __init__(self, ord=2):
        self.ord = 2

    def compute_shape(self, *dims: Dimension) -> Dimension:
        return Dimension(None)

    def is_linear(self):
        return False


class ClipOP(Operator):
    def __init__(self, a=None, a_min=None, a_max=None):
        self.lower = a_min
        self.higher = a_max
        _ = a

    def pre_process(self, *args):
        if len(args) == 1:
            return args[0]

        a, a_min, a_max = args
        self.lower = a_min
        self.higher = a_max
        return a

    def compute_shape(self, *dims: Dimension) -> Dimension:
        assert all(d.is_scalar() for d in dims)
        return Dimension(None)

    def is_linear(self):
        return False


compute_shape: Dict[OP, Callable[[Dimension, Dimension], Dimension]] = {}


def register_shape(*ops: OP):
    def inner(func):
        for op in ops:
            compute_shape[op] = func

        return func

    return inner


@register_shape(OP.EVALUATE)
def evaluate_shape(function_sig: FunctionSpace, *args: Dimension):

    if len(args) != len(function_sig.arguments):
        raise InvalidShape(
            f"Expected {len(function_sig.arguments)} arguments, got {len(args)}"
        )

    for i, (arg_dim, input_dim) in enumerate(
        zip(args, function_sig.input_dimensions())
    ):
        if arg_dim != input_dim:
            raise InvalidShape(
                f"Argument {i} has dimension {arg_dim.dim}, expected {input_dim.dim}"
            )
    try:
        (out_dim,) = function_sig.output_dimensions()
        return out_dim
    except ValueError:
        return function_sig.output_dimensions()


__componentwise_ops = [
    OP.NEG,
    OP.ABS,
    OP.SIN,
    OP.COS,
    OP.TAN,
    OP.ARCSIN,
    OP.ARCCOS,
    OP.ARCTAN,
    OP.EXP,
    OP.LOG,
]


@register_shape(OP.VALUE, *__componentwise_ops)
def dimension_identity(dim: Dimension):
    return dim


@register_shape(OP.EQUAL, OP.LESS_EQUAL, OP.LESS_THAN)
def comparison_dimension(dim1: Dimension, dim2: Dimension):
    if dim1.dim != dim2.dim:
        raise InvalidShape(dim1, dim2)
    return dim1


@register_shape(OP.CASE)
def case_dimension(
    condition: Dimension, false_branch: Dimension, true_branch: Dimension
):
    if not condition.is_scalar():
        raise InvalidShape("Condition must be a scalar")

    if false_branch != true_branch:
        raise InvalidShape(
            "Arguments are of different dimensions", false_branch, true_branch
        )

    return true_branch


@register_shape(OP.ADD, OP.SUB, OP.ARCTAN2)
def same_dimension(d_1: Dimension, d_2: Dimension) -> Dimension:
    if d_1.is_scalar() and d_2.is_scalar():
        return d_1

    if d_1 != d_2:
        raise InvalidShape("Arguments are of different dimensions", d_1, d_2)

    return d_1


@register_shape(OP.MUL, OP.DIV)
def shape_mul(d_1: Dimension, d_2: Dimension):
    if d_1.is_scalar():
        return d_2

    if d_2.is_scalar():
        return d_1
    if d_1 == d_2:
        return d_1

    raise InvalidArgument(
        "Multiplication is not define between two non-scalars. "
        "Consider using other operations",
        d_1,
        d_2,
    )


@register_shape(OP.MATMUL)
def shape_matmul(d_1: Dimension, d_2: Dimension):

    if d_1.is_scalar() or d_2.is_scalar():
        raise InvalidArgument(
            "Matrix multiplication is not defined for scalars"
        )

    if d_1.is_vector():
        raise InvalidArgument("Cannot multiply vectors")

    c = d_1.dim[-1]
    out_dims = []
    # if d_1.is_matrix() or d_1.is_multilinear_map():
    try:
        out_dims += [d for d in d_1.dim[:-1]]
    except IndexError:
        pass
    if isinstance(d_2, FunctionSpace):
        assert len(d_2.output_dimensions()) == 1
        d_2 = d_2.output_dimensions()[0]
    r = d_2.dim[0]

    #    if d_2.is_matrix() or d_2.is_multilinear_map():
    try:
        out_dims += [d for d in d_2.dim[1:]]
    except IndexError:
        pass

    if c != r:
        raise InvalidArgument(
            "Cannot multiply: product axis has different shape",
            d_1.shape,
            d_2.shape,
        )

    if out_dims:
        try:
            return Dimension(tuple(out_dims))
        except TypeError as ex:
            raise ex
    else:
        return Dimension(None)


@register_shape(OP.EXP, OP.SQRT)
def scalar_shape(d: Dimension):
    if d.is_scalar():
        return d
    raise InvalidArgument("Can only operate on scalar dimensions.")


@register_shape(OP.CROSS)
def cross_shape(d_1: Dimension, d_2: Dimension):
    if d_1.dim != d_2.dim != (3,):
        raise InvalidArgument("Cross product is only defined for (3,) vectors")
    return Dimension((3,))


@register_shape(OP.DOT)
def dot_shape(d_1: Dimension, d_2: Dimension):
    if d_1.dim == d_2.dim and (d_1.is_vector() or d_1.is_covector()):
        return Dimension(None)

    raise InvalidArgument(
        "Dot product only defined for vectors from the same space."
    )


@register_shape(OP.TRANSPOSE)
def transpose_shape(d: Dimension):
    if d.is_scalar():
        return d

    if d.is_vector():
        return Dimension((1, d.dim[0]))
    if d.is_covector():
        return Dimension((d.dim[1],))

    return Dimension(tuple(reversed(d.dim)))


numpy_atomics = {
    np.cross: OP.CROSS,
    np.dot: OP.DOT,
    np.matmul: OP.MATMUL,
    np.add: OP.ADD,
    np.multiply: OP.MUL,
    np.subtract: OP.SUB,
    np.sin: OP.SIN,
    np.cos: OP.COS,
    np.tan: OP.TAN,
    np.exp: OP.EXP,
    np.power: OP.PWR,
    np.transpose: OP.TRANSPOSE,
    np.divide: OP.DIV,
    np.arccos: OP.ARCCOS,
    np.arcsin: OP.ARCSIN,
    np.sqrt: OP.SQRT,
    np.negative: OP.NEG,
    np.abs: OP.ABS,
    np.arctan2: OP.ARCTAN2,
    np.log: OP.LOG,
}


numpy_composites = {
    np.concatenate: ConcatenateOP,
    np.reshape: ReshapeOP,
    np.linalg.norm: NormOP,
    np.clip: ClipOP,
}



import numpy as np

from coker import normalise
from coker.toolkits.spatial.types import Vec3, Scalar
from coker.algebra.kernel import Tracer


def quaternion_mul(q, p):

    qp_0 = q.q_0 * p.q_0 - np.dot(q.v, p.v)
    qp_v = q.q_0 * p.v + p.q_0 * q.v + np.cross(q.v, p.v)

    return UnitQuaternion(qp_0, qp_v)


class UnitQuaternion:
    __slots__ = ["q_0", "v"]

    def __str__(self):
        return f"UnitQuaternion({self.q_0}, {self.v})"

    def __init__(self, q_0: Scalar, v: Vec3):
        self.q_0 = q_0
        self.v = v

    def __eq__(self, other):
        return (
            isinstance(other, UnitQuaternion)
            and self.q_0 == other.q_0
            and (self.v == other.v).all()
        )

    @staticmethod
    def from_axis_angle(axis: Vec3, angle: Scalar):
        try:
            if not axis.nonzero():
                return UnitQuaternion(1, np.array([0, 0, 0]))
        except AttributeError:
            pass
        q_0 = np.cos(angle / 2)
        s = np.sin(angle / 2)

        u, _ = normalise(axis)

        v = s * u
        return UnitQuaternion(q_0, v)

    def __mul__(self, other) -> "UnitQuaternion":
        if isinstance(other, UnitQuaternion):
            return quaternion_mul(self, other)

        assert other.shape == (
            3,
        ), f"Cannot mul a vector of shape {other.shape}"
        p = UnitQuaternion(0, other)

        return quaternion_mul(self, p)

    def __rmul__(self, other):
        if isinstance(other, Vec3):
            assert other.shape == (3,)
            p = UnitQuaternion(0, other)
        elif isinstance(other, UnitQuaternion):
            p = other
        else:
            raise NotImplementedError(
                "Can't right multiply by {}".format(other)
            )

        return quaternion_mul(p, self)

    def inverse(self):
        return UnitQuaternion(self.q_0, -self.v)

    def conjugate(self, other):
        if isinstance(other, UnitQuaternion):
            return self * other * self.inverse()
        elif isinstance(other, Vec3):
            qpq_inv = self * other * self.inverse()
            return qpq_inv.v
        elif isinstance(other, Tracer):
            assert other.shape[0] == 3
            qpq_inv = self * other * self.inverse()
            result = qpq_inv.v
            return np.reshape(result, shape=other.shape)

        raise NotImplementedError(
            f"Quaternion conjugation not implemented for {type(other)}"
        )

    @staticmethod
    def from_elements(q_0, q_1, q_2, q_3):
        return UnitQuaternion(q_0, np.array([q_1, q_2, q_3]))

    def to_elements(self):
        return [self.q_0, self.v[0], self.v[1], self.v[2]]



from typing import List, Tuple
from coker.toolkits.system_modelling.modelling.coker_abc import (
    CokerListableSubclasses,
)


class ComponentHint:
    pass


def list_components() -> List[Tuple[str, str, ComponentHint]]:
    """List all components

    Returns:
        list of  component name, library path, usage hint

    """
    entries = []
    for baseclass in CokerListableSubclasses.list_subclasses():
        for component in baseclass.list_subclasses():
            path = component.__module__.replace(".", "/")

            entries.append((component.__name__, path, None))

    return entries



import casadi as ca
import numpy as np

import coker
from coker import OP, Tape, Tracer, FunctionSpace, Noop
from coker.algebra.ops import ConcatenateOP, ReshapeOP, NormOP
from typing import List

impls = {
    OP.ADD: lambda x, y: x + y,
    OP.SUB: lambda x, y: x - y,
    OP.MUL: lambda x, y: x * y,
    OP.DIV: lambda x, y: x / y,
    OP.MATMUL: lambda x, y: x @ y,
    OP.SIN: ca.sin,
    OP.COS: ca.cos,
    OP.TAN: ca.tan,
    OP.EXP: ca.exp,
    OP.PWR: ca.power,
    OP.INT_PWR: ca.power,
    OP.ARCCOS: ca.arccos,
    OP.ARCSIN: ca.arcsin,
    OP.DOT: ca.dot,
    OP.CROSS: ca.cross,
    OP.TRANSPOSE: ca.transpose,
    OP.NEG: lambda x: -x,
    OP.SQRT: ca.sqrt,
    OP.ABS: ca.fabs,
    OP.ARCTAN2: ca.atan2,
    OP.EQUAL: ca.eq,
    OP.LESS_EQUAL: ca.le,
    OP.LESS_THAN: ca.lt,
    OP.CASE: lambda c, t, f: ca.if_else(c, t, f),
    OP.EVALUATE: lambda op, *args: casadi_eval(op, *args),
    OP.LOG: ca.log,
}


def casadi_eval(op, *args):
    # Plain Python callables (e.g. solution_proxy) can be called directly.
    if not hasattr(op, "tape"):
        return op(*args)
    # coker.Function: evaluate symbolically via substitute, keeping CasADi
    # types (MX/DM) throughout rather than converting to Python scalars.
    workspace = {
        idx: arg for idx, arg in zip(op.tape.input_indicies, args) if idx >= 0
    }
    result = substitute(op.output, workspace)
    if op.is_single:
        return result[0]
    return result


def concat(*args: ca.MX, axis=0):
    if not axis:
        return ca.vertcat(*args)

    if axis == 1:
        return ca.horzcat(*args)

    # axis = 0 -> vstack
    # axis = 1 -> hstack
    # axis = None -> flatten +
    raise NotImplementedError


def norm(x, ord):
    if ord == 1:
        return ca.norm_1(x)
    if ord == 2:
        return ca.norm_2(x)

    raise NotImplementedError


def reshape(x, *shape):
    if len(shape) == 1:
        return ca.reshape(x, *shape, 1)
    else:
        return ca.reshape(x, *shape)


parameterised_impls = {
    ConcatenateOP: lambda op, *args: concat(*args, axis=op.axis),
    NormOP: lambda op, x: norm(x, ord=op.ord),
    ReshapeOP: lambda op, x: reshape(x, *op.newshape),
}


def call_parameterised_op(op, *args):
    kls = op.__class__
    try:
        result = parameterised_impls[kls](op, *args)
    except KeyError as ex:
        raise KeyError(f"Operation {op} not implemented.") from ex

    return result


class CasadiTensor:
    def __init__(self, *shape):
        self.shape = shape
        self.data = {}

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        return 0

    def __mul__(self, other):
        self.data = {k: d * other for k, d in self.data.items()}

    def __matmul__(self, other):
        assert (
            len(self.shape) == 3
        ), f"Higher order tensors not yet implemented"
        assert len(other.shape) == 2
        assert other.shape[0] == self.shape[-1]
        assert other.shape[1] == 1

        out = ca.MX(self.shape[0], self.shape[1])

        for (i, j, k), v in self.data.items():
            out[i, j] += v * other[k, 0]

        return out

    def reshape(self, shape):
        assert self.shape == shape
        return self


def to_casadi(value):
    import scipy.sparse

    if scipy.sparse.issparse(value):
        return ca.DM(scipy.sparse.csc_matrix(value))
    if isinstance(value, np.ndarray):
        if len(value.shape) == 1:
            value = value.reshape(-1, 1)

        if len(value.shape) > 2:
            v = CasadiTensor(*value.shape)
        else:
            v = ca.DM.zeros(*value.shape)
        it = np.nditer(value, op_flags=["readonly"], flags=["multi_index"])
        for x in it:
            if x != 0:
                k = it.multi_index
                v[k] = x
        return v
    try:
        if value == np.inf:
            return ca.inf
        if value == -np.inf:
            return -ca.inf
    except RuntimeError:
        pass

    return value


def extract_symbols(arg: ca.MX):
    if isinstance(arg, (ca.Function, coker.Function)):
        return set()
    v = {arg.dep(i) for i in range(arg.n_dep()) if arg.dep(i).is_symbolic()}
    return v


def substitute(output: List[Tracer], workspace):
    def get_node(node: Tracer):
        if node is None or node.index == Tape.NONE:
            return None
        if node is Noop():
            return node
        if node.index in workspace:
            return workspace[node.index]

        if node.is_constant():
            v = to_casadi(node.value())
        else:
            op, *args = node.value()
            args = [get_node(a) for a in args]

            if op in impls:
                try:
                    v = impls[op](*args)
                except RuntimeError as e:
                    raise e
            else:
                v = call_parameterised_op(op, *args)
        try:
            if not node.dim.is_scalar():
                shape = (
                    node.shape
                    if not node.dim.is_vector()
                    else (*node.dim.shape, 1)
                )
                v = v.reshape(shape)
        except AttributeError:
            pass

        workspace[node.index] = v
        return v

    return [get_node(o) for o in output]


def lower(tape: Tape, output: List[Tracer], workspace=None):
    workspace = {} if not workspace else workspace
    inputs = dict()
    for i in tape.input_indicies:
        if i in workspace:
            s = extract_symbols(workspace[i])
            inputs.update({s_i.__hash__(): s_i for s_i in s})
            continue

        assert not isinstance(
            tape.dim[i], FunctionSpace
        ), "Cannot lower a partially evaluated function."

        v = ca.MX.sym(f"x_{i}", *tape.dim[i].shape)
        workspace[i] = v
        inputs[v.__hash__()] = v

    result = substitute(output, workspace)
    return list(inputs.values()), result



import numpy as np


class SymbolicVector:
    def __init__(self, shape, data=None, symbolics=None):

        self.shape = shape
        self.dense_consts = data if data is not None else np.zeros(shape)
        self.symbolics = symbolics if symbolics is not None else {}

    @staticmethod
    def from_array(array: np.ndarray):

        assert array.dtype != np.object_, f"{array}"
        return SymbolicVector(array.shape, array)

    @staticmethod
    def zeros(shape):
        return SymbolicVector(shape)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self._set_row(key, value)
        else:
            raise NotImplementedError

    def __add__(self, other):
        if not self.symbolics:
            return self.dense_consts + other
        else:
            raise NotImplementedError

    def __sub__(self, other):
        if not self.symbolics:
            return self.dense_consts - other
        else:
            raise NotImplementedError

    def __mul__(self, other):
        if not self.symbolics:
            return self.dense_consts * other
        else:
            raise NotImplementedError

    def __rmul__(self, other):
        if not self.symbolics:
            return other * self.dense_consts
        else:
            raise NotImplementedError

    def _set_row(self, index: int, value):
        if isinstance(value, (float, int)):
            self.dense_consts[index] = value
        else:
            self.symbolics[index] = value

    def collapse(self):
        out = self.dense_consts
        for key, value in self.symbolics.items():
            p = projector_from_key(self.shape, key)
            out += value * p

        return out

    @staticmethod
    def from_list(data):
        shape = (len(data),)
        dense = np.zeros(shape)
        symbols = {}

        for row, val in enumerate(data):

            if isinstance(val, (float, int)):
                dense[row] = val

            elif isinstance(val, (list, tuple)):
                raise NotImplementedError
            else:
                symbols[row] = val

        if not symbols:
            return dense
        else:
            return SymbolicVector(shape, dense, symbols)


def projector_from_key(shape, key):
    assert isinstance(key, int)
    basis = np.zeros(shape)
    basis[key] = 1
    return basis



from coker.backends.coker.sparse_tensor import (
    dok_ndarray,
    tensor_vector_product,
)


def hat(a):
    return tensor_vector_product(levi_civita_3_tensor, a, axis=1)


levi_civita_3_tensor = dok_ndarray(
    (3, 3, 3),
    {
        (0, 1, 2): 1,
        (2, 0, 1): 1,
        (1, 2, 0): 1,
        (0, 2, 1): -1,
        (1, 0, 2): -1,
        (2, 1, 0): -1,
    },
)


def dot_tensor(n):
    return dok_ndarray((1, n, n), data={(0, i, i): 1 for i in range(n)})



from coker import Dimension, OP, scalar_types
from coker.backends import get_backend_by_name
from coker.backends.coker.sparse_tensor import dok_ndarray
from coker.backends.coker.memory import MemorySpec
from coker.backends.coker.weights import BilinearWeights, _CompiledBW

_AnyBW = (BilinearWeights, _CompiledBW)

from typing import List, Set, Tuple, Dict


import numpy as np


def vec(item):
    if isinstance(item, list):
        item = np.array(item)
    if isinstance(item, np.ndarray):
        return item.flatten(order="F")
    if isinstance(item, scalar_types):
        return np.array([item])
    raise NotImplementedError(type(item))


class InputLayer:
    def __init__(self):
        """maps such that arg_i = map_i(x)

        - For scalars; map_i is a linear operator
        - For vector; map_i is a matrix


        """
        self.vec_to_arg_maps: List[Set[Tuple[int, ...]]] = []
        self.out_shape = []
        self.dimension = 0

    def add_input(self, dim: Dimension) -> int:

        idx = len(self.vec_to_arg_maps)
        if dim.is_scalar():
            self.vec_to_arg_maps.append({(0, self.dimension)})
            self.dimension += 1
            self.out_shape.append((1,))
        else:
            basis = set()
            for i, idx in enumerate(dim.index_iterator(row_major=False)):
                # for example, a vector would have an index tuple
                # of idx = (row, ) so the basis becomes (row, d).
                # We interpret this as a matrix
                # For a matrix, we have bases (row, col, d)

                basis.add((*idx, self.dimension))
                self.dimension += 1
            self.vec_to_arg_maps.append(basis)
            self.out_shape.append((*dim,))

        return idx

    def get_projection(self, arg: int):
        # return a matrix that maps
        # M: vec(x) -> arg_i

        nzeros = self.vec_to_arg_maps[arg]
        shape = (*self.out_shape[arg], self.dimension)
        m = dok_ndarray(shape)
        for idx in nzeros:
            m[idx] = 1
        return m

    def __call__(self, *args):
        assert len(args) == len(self.vec_to_arg_maps)

        x = np.concatenate([vec(a) for a in args])
        return x

    def forwards(self, *tangent_space, y=None):
        r"""Reverse mode autodiff

        Here, :math:`y = \sum_j P_j x_j$`

        so  dy = \sum P_j dx_j

        """
        n_args = len(self.vec_to_arg_maps)
        assert len(tangent_space) == 2 * n_args
        dx = tangent_space[n_args : 2 * n_args]

        return np.concatenate([vec(dx_i) for dx_i in dx])


def to_float(x):
    if isinstance(x, (float, int)):
        return float(x)

    assert isinstance(x, np.ndarray)
    x_out = x.flatten()
    assert x_out.shape == (1,)
    return x_out[0]


class OutputLayer:

    def __init__(self):
        self.outputs = {}

    def inputs(self):
        return list(self.outputs.keys())

    def add_output(self, memory: MemorySpec, shape: Dimension):
        assert memory not in self.outputs
        if shape.dim is None:
            self.outputs[memory] = to_float
            return

        def shape_fn(array):
            return np.reshape(array, shape.dim)

        self.outputs[memory] = shape_fn

    def call(self, context: Dict[MemorySpec, np.ndarray]):
        result = [f(context[k]) for k, f in self.outputs.items()]
        if len(result) == 1:
            return result[0]
        return result


class GenericLayerOP:
    def __init__(self, output, op, *weights):
        self.op = op
        self.weights = weights
        self.output = output

        assert all(
            isinstance(w, (*_AnyBW, np.ndarray, int, float, bool))
            for w in weights
        ), f"Unkown type in {weights}"

    def inputs(self) -> List[MemorySpec]:
        return [w.memory for w in self.weights if isinstance(w, _AnyBW)]

    def outputs(self) -> List[MemorySpec]:
        return [self.output]

    def __call__(self, *x):
        backend = get_backend_by_name("numpy", set_current=False)
        evaluated = []
        xi = iter(x)
        for w in self.weights:
            if isinstance(w, _AnyBW):
                evaluated.append(w(next(xi)))
            else:
                evaluated.append(w)
        return backend.call(self.op, *evaluated)

    def push_forward(self, *tangent_space):
        n_bw = sum(1 for w in self.weights if isinstance(w, _AnyBW))
        x_bw_vals, dx_bw_vals = tangent_space[0:n_bw], tangent_space[n_bw:]
        y = self(*x_bw_vals)

        bw_iter = iter(zip(x_bw_vals, dx_bw_vals))
        x_eval = []
        dx_eval = []
        for weight in self.weights:
            if isinstance(weight, _AnyBW):
                x_i, dx_i = next(bw_iter)
                x_out, dx_out = weight.push_forwards(x_i, dx_i)
                x_eval.append(x_out)
                dx_eval.append(dx_out)
            else:
                constant = (
                    weight
                    if isinstance(weight, np.ndarray)
                    else np.array([weight])
                )
                x_eval.append(constant)
                dx_eval.append(np.zeros_like(constant))

        df = differentials[self.op](*x_eval)
        if all(isinstance(df_i, (dok_ndarray, np.ndarray)) for df_i in df):
            dy = sum(df_i @ dx_i for df_i, dx_i in zip(df, dx_eval))
        else:
            dy = sum(df_i * dx_i for df_i, dx_i in zip(df, dx_eval))
        return y, dy


def d_dot(x, y) -> Tuple:
    # d = x.T @ y  = sum(x_iy_i)
    # dd_i = y.T @ dx_i
    # dd_i = x.T @ dy_i
    return y.T, x.T


differentials = {
    OP.SIN: np.cos,
    OP.COS: lambda x: -np.sin(x),
    OP.MUL: lambda a, b: [b, a],
    OP.ADD: lambda a, b: [1, 1],
    OP.SUB: lambda a, b: [1, -1],
    OP.DOT: d_dot,
}
